# -*- coding: utf-8 -*-
"""
Quality control module for ObsPy.

Currently requires MiniSEED files as that is the dominant data format in
data centers. Please see the documentation of the
:class:`~obspy.signal.quality_control.MSEEDMetadata` class for usage
instructions.

:authors:
    Luca Trani (trani@knmi.nl)
    Lion Krischer (krischer@geophysik.uni-muenchen.de)
    Mathijs Koymans (koymans@knmi.nl)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import collections.abc
import io
import json
from operator import attrgetter
from pathlib import Path
from uuid import uuid4

import numpy as np

from obspy import Stream, UTCDateTime, read, __version__
from obspy.core.util.base import get_dependency_version
from obspy.io.mseed.util import get_flags


_PRODUCER = "ObsPy %s" % __version__


class DataQualityEncoder(json.JSONEncoder):
    """
    Custom encoder capable of dealing with NumPy and ObsPy types.
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, UTCDateTime):
            return str(obj)
        else:
            return super(DataQualityEncoder, self).default(obj)


class MSEEDMetadata(object):
    """
    A container for MiniSEED specific metadata, including quality control
    parameters.

    Reads the MiniSEED files and extracts the data quality metrics. All
    MiniSEED files must have a matching stream ID and quality.

    :param files: One ore more MiniSEED files.
    :type files: str or list[str]
    :type id: str, optional
    :param id: A unique identifier of the to be created QC object. It is
        not verified, that it actually is unique. The user has to take care of
        that. If no id is given, uuid.uuid4() will be used to
        create one which assures uniqueness within one Python run.
        If no fixed id is provided, the ID will be built from prefix
        and a random uuid hash.
    :type prefix: str, optional
    :param prefix: An optional identifier that will be put in front of any
        automatically created id. The prefix will only have an effect
        if `id` is not specified (for a fixed ID string).
    :param starttime: Only use records whose end time is larger then this
        given time. Also specifies the new official start time of the
        metadata object.
    :type starttime: :class:`obspy.core.utcdatetime.UTCDateTime`
    :param endtime: Only use records whose start time is smaller then this
        given time. Also specifies the new official end time of the
        metadata object
    :type endtime: :class:`obspy.core.utcdatetime.UTCDateTime`
    :param add_c_segments: Calculate metrics for each continuous segment.
    :type add_c_segments: bool
    :param add_flags: Include MiniSEED header statistics in result.
    :type add_flags: bool
    :param waveform_type: The type of waveform data, e.g. ``"seismic"``,
        ``"infrasound"``, ...
    :type waveform_type: str

    .. rubric:: Example

    >>> from obspy.signal.quality_control import
    ...     MSEEDMetadata #doctest: +SKIP
    >>> mseedqc = MSEEDMetadata(['path/to/file',
    ...                          'path/to/file2']) # doctest: +SKIP

    The class requires a list of files for calculating metrics.
    Add optional parameters ``starttime="YYYY-MM-DDThh:mm:ss`` and
    ``endtime="YYYY-MM-DDThh:mm:ss"`` or ``obspy.core.utcdatetime.UTCDateTime``
    to limit metric calculation to this window. Continuous segments are
    returned when ``add_c_segments=True`` and MiniSEED header flags information
    is returned when ``add_flags=True``.

    The calculated metrics are then available in the ``.meta`` dictionary.

    >>> mseedqc.meta  # doctest: +SKIP

    This is intended to be serialized as JSON. Retrieve the JSON string (to
    for example store it in a database or save to a file) with:

    >>> mseedqc.get_json_meta() #doctest: +SKIP
    """
    def __init__(self, files, id=None, prefix="smi:local/qc",
                 starttime=None, endtime=None,
                 add_c_segments=True, add_flags=False,
                 waveform_type="seismic"):
        """
        Reads the MiniSEED files and extracts the data quality metrics.
        """
        self.data = Stream()
        self.all_files = files
        self.files = []

        # Allow anything UTCDateTime can parse.
        if starttime is not None:
            starttime = UTCDateTime(starttime)
        if endtime is not None:
            endtime = UTCDateTime(endtime)

        self.window_start = starttime
        self.window_end = endtime

        # We are required to exclude samples at T1. Therefore, shift the
        # time window to the left by 1μs and set nearest_sample to False.
        # This will force ObsPy to fall back to the sample left of the endtime
        if endtime is not None:
            endtime_left = endtime - 1e-6
        else:
            endtime_left = None

        # Will raise if not a MiniSEED files.
        for file in files:
            st = read(file, starttime=starttime, endtime=endtime_left,
                      format="mseed", nearest_sample=False)

            # Empty stream or maybe there is no data in the stream for the
            # requested time span.
            if not st:
                continue

            self.files.append(file)

            # Only extend traces with data (npts > 0)
            for tr in st:
                if tr.stats.npts != 0:
                    self.data.extend([tr])

        if not self.data:
            raise ValueError("No data within the temporal constraints.")

        # Do some sanity checks. The class only works with data from a
        # single location so we have to make sure that the existing data on
        # this object and the newly added all have the same identifier.
        ids = [tr.id + "." + str(tr.stats.mseed.dataquality) for tr in
               self.data]
        if len(set(ids)) != 1:
            raise ValueError("All traces must have the same SEED id and "
                             "quality.")

        # Get the last sample and add delta
        final_trace = max(self.data, key=attrgetter('stats.endtime')).stats
        self.endtime = endtime or final_trace.endtime + final_trace.delta

        self.data.sort()

        # Set the metric start and endtime specified by the user.
        # If no start and endtime are given, we pick our own, and the window
        # will start on the first sample and end on the last sample + Δt.
        # This is conform to the definition of [T0, T1).
        self.starttime = starttime or self.data[0].stats.starttime
        self.total_time = self.endtime - self.starttime

        if id is None:
            id = prefix + "/" + str(uuid4())

        # Fill with the meta information.
        self.meta = {
            "wfmetadata_id": id,
            "producer": _PRODUCER,
            "waveform_type": waveform_type,
            "waveform_format": "miniSEED",
            "version": "1.0.0"
        }

        # Get sample left of the user specified starttime
        # This will allow us to determine start continuity in our window
        self._get_gaps_and_overlaps()

        # The calculation of all the metrics begins here
        self._extract_mseed_stream_metadata()
        self._compute_sample_metrics()

        if add_flags:
            self._extract_mseed_flags()

        if add_c_segments:
            self._compute_continuous_seg_sample_metrics()

    def _get_gaps_and_overlaps(self):
        """
        Function to get all gaps and overlaps in the user
        specified (or forced) window.
        """
        self.all_data = Stream()

        body_gap = []
        body_overlap = []

        # Read all the files entirely and calculate gaps and overlaps
        # for the entire segment. Later we will narrow it to our window if
        # it has been specified
        for file in self.all_files:
            self.all_data.extend(read(file, format="mseed", headonly=True))

        # Sort the data by so the start times are in order
        self.all_data.sort()

        # Coverage keeps track of the time used
        coverage = None
        for trace in self.all_data:

            # Extend the endtime of a trace with delta
            trace_end = trace.stats.endtime + trace.stats.delta
            trace_start = trace.stats.starttime

            # If a start boundary has been specified
            if self.window_start is not None:
                if trace_end <= self.window_start:
                    continue

                # In case the start time of a trace comes before the window
                # extend the length to the window start
                cut_trace_start = max(trace_start, self.window_start)
            else:
                cut_trace_start = trace_start

            # If a end boundary has been specified
            if self.window_end is not None:
                if trace_start > self.window_end:
                    continue

                # In case the end time of a trace comes after the window
                # reduce the length to the window end
                cut_trace_end = min(trace_end, self.window_end)
            else:
                cut_trace_end = trace_end

            # Calculate the trace time tolerance as 0.5 * delta
            time_tolerance_max = trace_end + 0.5 * trace.stats.delta
            time_tolerance_min = trace_end - 0.5 * trace.stats.delta

            # Set the initial trace coverage and proceed to the next trace
            if coverage is None:
                coverage = {
                    'start': trace_start,
                    'end': trace_end,
                    'end_min': time_tolerance_min,
                    'end_max': time_tolerance_max
                }
                continue

            # Check if the start time of a trace falls
            # beyond covered end max (time tolerance)
            # this must be interpreted as a gap
            if trace_start > coverage['end_max']:
                body_gap.append(cut_trace_start - coverage['end'])

            # Check overlap in coverage
            # the overlap ends at the end of the trace
            # or add the end of the coverage
            if trace_start <= coverage['end_min']:
                min_end = min(cut_trace_end, coverage['end'])
                body_overlap.append(min_end - cut_trace_start)

            # Extend the coverage of the trace
            # the start coverage remains unchanged
            if trace_end > coverage['end']:
                coverage['end'] = trace_end
                coverage['end_min'] = time_tolerance_min
                coverage['end_max'] = time_tolerance_max

        # We have the coverage by the traces
        # check if there is an end or start gap caused by the
        # window forced by the user
        self.meta['start_gap'] = None
        self.meta['end_gap'] = None
        if self.window_start is not None:
            if coverage['start'] > self.window_start:
                self.meta['start_gap'] = coverage['start'] - self.window_start
                body_gap.append(self.meta['start_gap'])
        if self.window_end is not None:
            if coverage['end'] < self.window_end:
                self.meta['end_gap'] = self.window_end - coverage['end']
                body_gap.append(self.meta['end_gap'])

        # Set the gap and overlap information
        self.meta['num_gaps'] = len(body_gap)
        self.meta['sum_gaps'] = sum(body_gap)
        if len(body_gap) != 0:
            self.meta['max_gap'] = max(body_gap)
        else:
            self.meta['max_gap'] = None

        self.meta['num_overlaps'] = len(body_overlap)
        self.meta['sum_overlaps'] = sum(body_overlap)
        if len(body_overlap) != 0:
            self.meta['max_overlap'] = max(body_overlap)
        else:
            self.meta['max_overlap'] = None

    @property
    def number_of_records(self):
        """
        Number of records across files before slicing.
        """
        return sum(tr.stats.mseed.number_of_records for tr in self.data)

    @property
    def number_of_samples(self):
        """
        Number of samples across files.
        """
        return sum(tr.stats.npts for tr in self.data)

    def _extract_mseed_stream_stats(self):
        """
        Small function to collects the mSEED stats
        """
        stats = self.data[0].stats
        self.meta['network'] = stats.network
        self.meta['station'] = stats.station
        self.meta['location'] = stats.location
        self.meta['channel'] = stats.channel
        self.meta['quality'] = stats.mseed.dataquality

    def _extract_mseed_stream_metadata(self):
        """
        Collect information from the MiniSEED headers.
        """
        self._extract_mseed_stream_stats()

        meta = self.meta

        # Save first and last sample of the trace
        # Look for the maximum endtime and minimum starttime in case
        # traces are not in order.
        meta['first_sample'] = min(tr.stats.starttime for tr in self.data)
        meta['last_sample'] = max(tr.stats.endtime for tr in self.data)

        # Add some other parameters to the metadata object
        meta['seed_id'] = self.data[0].id
        meta['files'] = self.files
        meta['start_time'] = self.starttime
        meta['end_time'] = self.endtime
        meta['num_samples'] = self.number_of_samples

        # The number records as given by Trace.stats is always
        # the full number of records in a file, regardless of being
        # sliced between a start & endtime
        # If a start/endtime is specified, we cannot be sure of
        # the # records. We will add this parameter later
        # if add_flags is set to true after looping all records
        if self.window_start is None and self.window_end is None:
            meta['num_records'] = self.number_of_records
        else:
            meta['num_records'] = None

        # The following are lists and may contain multiple unique entries.
        meta['sample_rate'] = \
            sorted(list(set([tr.stats.sampling_rate for tr in self.data])))
        meta['record_length'] = \
            sorted(list(set([tr.stats.mseed.record_length
                             for tr in self.data])))
        meta['encoding'] = \
            sorted(list(set([tr.stats.mseed.encoding for tr in self.data])))

    def _extract_mseed_flags(self):
        flags = get_flags(self.files, starttime=self.starttime,
                          endtime=self.endtime)

        data_quality_flags_seconds = flags["data_quality_flags_percentages"]
        activity_flags_seconds = flags["activity_flags_percentages"]
        io_and_clock_flags_seconds = flags["io_and_clock_flags_percentages"]
        timing_correction = flags["timing_correction"]

        # Only calculate the timing quality statistics if each files has the
        # timing quality set. This should usually be the case. Otherwise we
        # would created tinted statistics. There is still a chance that some
        # records in a file have timing qualities set and others not but
        # that should be small.
        if flags["timing_quality"]:
            tq = flags["timing_quality"]
            timing_quality_mean = tq["mean"]
            timing_quality_min = tq["min"]
            timing_quality_max = tq["max"]
            timing_quality_median = tq["median"]
            timing_quality_lower_quartile = tq["lower_quartile"]
            timing_quality_upper_quartile = tq["upper_quartile"]
        else:
            timing_quality_mean = None
            timing_quality_min = None
            timing_quality_max = None
            timing_quality_median = None
            timing_quality_lower_quartile = None
            timing_quality_upper_quartile = None

        meta = self.meta
        meta['num_records'] = flags['record_count']

        # Set MiniSEED header counts
        meta['miniseed_header_counts'] = {}
        ref = meta['miniseed_header_counts']
        ref['timing_correction'] = flags['timing_correction_count']
        ref['activity_flags'] = flags['activity_flags_counts']
        ref['io_and_clock_flags'] = flags['io_and_clock_flags_counts']
        ref['data_quality_flags'] = flags['data_quality_flags_counts']

        # Set MiniSEED header percentages
        meta['miniseed_header_percentages'] = {}
        ref = meta['miniseed_header_percentages']
        ref['timing_correction'] = timing_correction
        ref['timing_quality_mean'] = timing_quality_mean
        ref['timing_quality_min'] = timing_quality_min
        ref['timing_quality_max'] = timing_quality_max
        ref['timing_quality_median'] = timing_quality_median
        ref['timing_quality_lower_quartile'] = timing_quality_lower_quartile
        ref['timing_quality_upper_quartile'] = timing_quality_upper_quartile

        # According to schema @ maybe refactor this to less verbose flag
        # names. Sets MiniSEED header flag percentages
        ref['activity_flags'] = activity_flags_seconds
        ref['data_quality_flags'] = data_quality_flags_seconds
        ref['io_and_clock_flags'] = io_and_clock_flags_seconds

    def _compute_sample_metrics(self):
        """
        Computes metrics on samples contained in the specified time window
        """

        #
        # Vectorised implementation to reduce memory consumption
        # and optimise running time
        #

        full_samples = np.concatenate(
            [tr.data.astype(np.float64, copy=False) for tr in self.data])

        # Manually implement these as they have to work across a list of
        # arrays.
        self.meta['sample_mean'] = full_samples.mean()

        _min, _lower_quartile, _median, _upper_quartile, _max = \
            np.percentile(full_samples, [0, 25, 50, 75, 100])
        self.meta['sample_min'] = _min
        self.meta['sample_lower_quartile'] = _lower_quartile
        self.meta['sample_median'] = _median
        self.meta['sample_upper_quartile'] = _upper_quartile
        self.meta['sample_max'] = _max

        squared = np.mean(full_samples ** 2)
        self.meta['sample_rms'] = np.sqrt(squared)

        # Sample standard deviation
        self.meta['sample_stdev'] = \
            np.sqrt(squared - self.meta['sample_mean']**2)

        # Percentage based availability as a function of total gap length
        # over the full trace duration
        self.meta['percent_availability'] = 100 * (
            (self.total_time - self.meta['sum_gaps']) /
            self.total_time)

    def _compute_continuous_seg_sample_metrics(self):
        """
        Computes metrics on the samples within each continuous segment.
        """
        if not self.data:
            return

        if self.meta['start_gap'] is None and self.window_start is not None:
            first_segment_start = self.window_start
        else:
            first_segment_start = self.data[0].stats.starttime

        # Collect data in arrays of continuous segments
        # Manually set the first segment
        c_seg = {
            'start': first_segment_start,
            'data': self.data[0].data
        }

        c_segs = []
        for i in range(len(self.data)):

            trace_end = self.data[i].stats.endtime + self.data[i].stats.delta
            time_tolerance = 0.5 * self.data[i].stats.delta

            c_seg['s_rate'] = self.data[i].stats.sampling_rate
            c_seg['end'] = trace_end

            # Final trace, make sure to append
            if i == len(self.data) - 1:
                c_segs.append(c_seg)
                break

            trace_offset = abs(self.data[i + 1].stats.starttime - trace_end)

            # Check if trace endtime is equal to (trace + 1) starttime
            # and if the sampling_rates match, if so, extend the data with
            # data from trace + 1 and extend the endtime
            # Otherwise the segment stops being continuous and we append it
            # and we create a new data segment
            if (trace_offset < time_tolerance and
                    self.data[i + 1].stats.sampling_rate == c_seg['s_rate']):
                c_seg['data'] = np.concatenate((c_seg['data'],
                                                self.data[i + 1].data))
                c_seg['end'] = self.data[i + 1].stats.endtime + \
                    self.data[i + 1].stats.delta
            else:
                c_segs.append(c_seg)
                c_seg = {
                    'data': self.data[i + 1].data,
                    'start': self.data[i + 1].stats.starttime}

        # Set array of continuous segments from this data
        self.meta['c_segments'] = [self._parse_c_stats(seg) for seg in c_segs]

    def _parse_c_stats(self, tr):
        """
        :param tr: custom dictionary with start, end, data, and sampling_rate
            of a continuous trace
        """
        seg = {}

        # Set continuous segments start & end
        # limit to specified window start/end if set
        if self.window_start is not None:
            seg['start_time'] = max(self.window_start, tr['start'])
        else:
            seg['start_time'] = tr['start']

        if self.window_end is not None:
            seg['end_time'] = min(self.window_end, tr['end'])
        else:
            seg['end_time'] = tr['end']

        seg['sample_rate'] = tr['s_rate']
        seg['sample_min'] = tr['data'].min()
        seg['sample_max'] = tr['data'].max()
        seg['sample_mean'] = tr['data'].mean()
        seg['sample_median'] = np.median(tr['data'])
        seg['sample_rms'] = np.sqrt((tr['data'].astype(object) ** 2).sum() /
                                    len(tr['data']))
        seg['sample_lower_quartile'] = np.percentile(tr['data'], 25)
        seg['sample_upper_quartile'] = np.percentile(tr['data'], 75)
        seg['sample_stdev'] = tr['data'].std()
        seg['num_samples'] = len(tr['data'])
        seg['segment_length'] = seg['end_time'] - seg['start_time']

        return seg

    def get_json_meta(self, validate=False):
        """
        Serialize the meta dictionary to JSON.

        :param validate: Validate the JSON string against the schema before
            returning.
        :type validate: bool

        :return: JSON containing the MSEED metadata
        """
        meta = json.dumps(self.meta, cls=DataQualityEncoder, indent=4)

        if validate:
            self.validate_qc_metrics(meta)

        return meta

    def validate_qc_metrics(self, qc_metrics):
        """
        Validate the passed metrics against the JSON schema.

        :param qc_metrics: The quality metrics to be validated.
        :type qc_metrics: dict, str, or file-like object
        """
        import jsonschema

        # Judging from the changelog 1.0.0 appears to be the first version
        # to have fully working support for references.
        _v = get_dependency_version("jsonschema")
        if _v < [1, 0, 0]:  # pragma: no cover
            msg = ("Validating the QC metrics requires jsonschema >= 1.0.0 "
                   "You have %s. Please update." %
                   get_dependency_version("jsonschema", raw_string=True))
            raise ValueError(msg)

        schema_path = Path(__file__).parent/"data"/"wf_metadata_schema.json"
        with io.open(schema_path, "rt") as fh:
            schema = json.load(fh)

        # If passed as a dictionary, serialize and derialize to get the
        # mapping from Python object to JSON type.
        if isinstance(qc_metrics, collections.abc.Mapping):
            qc_metrics = json.loads(self.get_json_meta(validate=False))
        elif hasattr(qc_metrics, "read"):
            qc_metrics = json.load(qc_metrics)
        else:
            qc_metrics = json.loads(qc_metrics)

        jsonschema.validate(qc_metrics, schema)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



from obspy import read, read_inventory

st = read("/path/to/IU_ULN_00_LH1_2015-07-18T02.mseed")
tr = st[0]
inv = read_inventory("/path/to/IU_ULN_00_LH1.xml")
pre_filt = [0.001, 0.005, 10, 20]
tr.remove_response(inventory=inv, pre_filt=pre_filt, output="DISP",
                   water_level=60, plot=True)



# -*- coding: utf-8 -*-
"""
IASPEI Seismic Format (ISF) support for ObsPy

Currently only supports reading IMS1.0 bulletin files.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import warnings

from obspy import UTCDateTime
from obspy.core.event import (
    Catalog, Event, Origin, Comment, EventDescription, OriginUncertainty,
    QuantityError, OriginQuality, CreationInfo, Magnitude, ResourceIdentifier,
    Pick, StationMagnitude, WaveformStreamID, Amplitude, Arrival)
from obspy.core.event.header import AmplitudeUnit, OriginDepthType
from obspy.core.util.obspy_types import ObsPyReadingError
from obspy.io.iaspei.util import (
    float_or_none, int_or_none, fixed_flag, evaluation_mode_and_status,
    _block_header)


PICK_EVALUATION_MODE = {'m': 'manual', 'a': 'automatic', '_': None, '': None}
POLARITY = {'c': 'positive', 'd': 'negative', '_': None, '': None}
ONSET = {'i': 'impulsive', 'e': 'emergent', 'q': 'questionable', '_': None,
         '': None}
EVENT_TYPE_CERTAINTY = {
    "uk": (None, None),
    "de": ("earthquake", "known"),
    "fe": ("earthquake", "known"),
    "ke": ("earthquake", "known"),
    "se": ("earthquake", "suspected"),
    "kr": ("rock burst", "known"),
    "sr": ("rock burst", "suspected"),
    "ki": ("induced or triggered event", "known"),
    "si": ("induced or triggered event", "suspected"),
    "km": ("mining explosion", "known"),
    "sm": ("mining explosion", "suspected"),
    "kh": ("chemical explosion", "known"),
    "sh": ("chemical explosion", "suspected"),
    "kx": ("experimental explosion", "known"),
    "sx": ("experimental explosion", "suspected"),
    "kn": ("nuclear explosion", "known"),
    "sn": ("nuclear explosion", "suspected"),
    "ls": ("landslide", "known"),
    "": (None, None),
    }
LOCATION_METHODS = {'i': 'inversion', 'p': 'pattern recognition',
                    'g': 'ground truth', 'o': 'other', '': None}

# Convert GSE2 depth flag to ObsPy depth type
DEPTH_TYPES = {
    'f': OriginDepthType('operator assigned'),
    'd': OriginDepthType('constrained by depth phases'),
}


class ISFEndOfFile(StopIteration):
    pass


def _decode_if_possible(value, encoding="UTF-8"):
    try:
        return value.decode(encoding)
    except Exception:
        return value


class ISFReader(object):
    encoding = 'UTF-8'
    resource_id_prefix = 'smi:local'

    def __init__(self, fh, skip_orphan=True, origin_specific_to_comments=False,
                 _no_uuid_hashes=False, **kwargs):
        """
        :param fh: File handle to read from.
        :type fh: file-like object
        :param skip_orphan:
            If False, orphan phase blocks will be skipped
            (i.e. if no origin could be assigned to the phase block).
        :type skip_orphan: bool
        :param origin_specific_to_comments:
            If True, origin-specific data in orphan phase blocks
            will be stored as text in the Pick.comments attribute.
            This is against the spirit of QuakeML.
            Used to maintain backward compatibility.
        :type origin_specific_to_comments: bool
        :param _no_uuid_hashes:
            If True, resource identifiers will not
            get uuid hashes appended to them.
            This is mainly useful for testing purposes.
        :type _no_uuid_hashes: bool
        """
        self.lines = [_decode_if_possible(line, self.encoding).rstrip()
                      for line in fh.readlines()
                      if line.strip()]
        self.cat = Catalog()
        self._no_uuid_hashes = _no_uuid_hashes
        self.skip_orphan = skip_orphan
        self.origin_specific_to_comments = origin_specific_to_comments

    def deserialize(self):
        if not self.lines:
            raise ObsPyReadingError()
        datasection_flag = False
        while not self._next_line_type():
            # Skip the lines preceding the begin of data section
            # (if the bulletin is enveloped as an AutoDRM response message)
            # (BEGIN IMS1.0, MSG_TYPE DATA, ...)
            line = self._get_next_line()
            if line.upper().startswith('DATA_TYPE BULLETIN IMS1.0'):
                if 'LONG' in line.upper():
                    # warnings.warn('Only IMS1.0:SHORT is supported, '
                    #               'other subformats might not work')
                    datasection_flag = False
                else:
                    # if subformat is not specified, assume SHORT
                    datasection_flag = True
                break
        if not datasection_flag:
            msg = ("No data section of valid DATA_TYPE found.\n"
                   "Is this a valid IMS1.0 bulletin?")
            raise ObsPyReadingError(msg)
        try:
            self._deserialize()
        except ISFEndOfFile:
            pass
        return self.cat

    def _deserialize(self):
        line = self._get_next_line()
        catalog_description = line.strip()
        self.cat.description = catalog_description
        if not self.lines[0].lower().startswith('event'):
            raise ObsPyReadingError()
        # get next line stops the loop eventually, raising a controlled
        # exception
        while True:
            next_line_type = self._next_line_type()
            if next_line_type == 'event':
                # scope last event when a new one starts
                # avoid trying to do it when starting to read first event,
                # which results in index error
                if len(self.cat):
                    self.cat[-1].scope_resource_ids()
                self._read_event_header()
            elif next_line_type:
                self._process_block()
            else:
                raise ObsPyReadingError
        # scope last event when no new events are encountered anymore
        # in principle we should get rid of the above scoping and should just
        # scope every event at the end in a loop? not sure if this has negative
        # side effects so doing it like this for now, just fixing the
        # IndexError for first event above
        if len(self.cat):
            self.cat[-1].scope_resource_ids()

    def _construct_id(self, parts, add_hash=False):
        id_ = '/'.join([str(self.cat.resource_id)] + list(parts))
        if add_hash and not self._no_uuid_hashes:
            id_ = str(ResourceIdentifier(prefix=id_))
        return id_

    def _get_next_line(self):
        if not self.lines:
            raise ISFEndOfFile
        line = self.lines.pop(0)
        if line.startswith('STOP'):
            raise ISFEndOfFile
        return line

    def _read_event_header(self):
        line = self._get_next_line()
        event_id = self._construct_id(['event', line[6:14].strip()])
        region = line[15:80].strip()
        event = Event(
            resource_id=event_id,
            event_descriptions=[EventDescription(text=region,
                                                 type='region name')])
        self.cat.append(event)

    def _next_line_type(self):
        if not self.lines:
            raise ISFEndOfFile
        return _block_header(self.lines[0])

    def _process_block(self):
        if not self.cat:
            raise ObsPyReadingError
        line = self._get_next_line()
        block_type = _block_header(line)
        # read origins block
        if block_type == 'origins':
            self._read_origins()
            self._specify_preferred_origin()
        # read publications block
        elif block_type == 'bibliography':
            self._read_bibliography()
        # read magnitudes block
        elif block_type == 'magnitudes':
            self._read_magnitudes()
        # read phases block
        elif block_type == 'phases':
            self._read_phases()
        # unexpected block header line
        else:
            msg = ('Unexpected line while reading file (line will be '
                   'ignored):\n' + line)
            warnings.warn(msg)

    def _read_phases(self):
        event = self.cat[-1]
        origin = event.preferred_origin()
        line = self.lines[0].strip()
        # OrigID Phase and Phase Information Comment
        if line.startswith('(#OrigID'):
            origin_id = ResourceIdentifier(
                self._construct_id(['origin', line[8:].rstrip(") ").strip()]))
            origin = origin_id.get_referred_object()
        if origin:
            origin_id = origin.resource_id
        else:
            origin_id = None
            msg = ('Event: {}'.format(event.resource_id) + '\n'
                   'Phase block cannot be fully processed because '
                   'it does not have an origin assigned.\n'
                   'Please check if the bulletin is in accordance '
                   'with the IMS1.0/ISF standard.')
            warnings.warn(msg)

        while not self._next_line_type():
            line = self._get_next_line()
            if self.skip_orphan and not origin_id:
                # skip all phase lines if no origin can be assigned
                continue
            if line.strip().startswith('('):
                comment = self._parse_generic_comment(line)
                if len(event.picks):
                    event.picks[-1].comments.append(comment)
                continue
            pick, amplitude, station_magnitude, arrival = self._parse_phase(
                    line, origin_id,
                    self.origin_specific_to_comments and not origin_id)
            if (pick, amplitude, station_magnitude) == (None, None, None):
                continue
            event.picks.append(pick)
            if amplitude:
                event.amplitudes.append(amplitude)
            if station_magnitude:
                event.station_magnitudes.append(station_magnitude)
            if arrival:
                # The list of Arrival items is stored in the specified origin
                if origin:
                    origin.arrivals.append(arrival)
            continue

    def _read_origins(self):
        event = self.cat[-1]
        origins = []
        event_types_certainties = []
        # just in case origin block is at end of file, make sure the event type
        # routine below gets executed, even if next line is EOF at some point
        try:
            while not self._next_line_type():
                line = self._get_next_line()
                if line.strip().startswith('('):
                    origins[-1].comments.append(
                        self._parse_generic_comment(line))
                    continue
                origin, event_type, event_type_certainty = \
                    self._parse_origin(line)
                origins.append(origin)
                event_types_certainties.append(
                    (event_type, event_type_certainty))
                continue
        finally:
            # check event types/certainties for consistency
            event_types = set(type_ for type_, _ in event_types_certainties)
            event_types.discard(None)
            if len(event_types) == 1:
                event_type = event_types.pop()
                certainties = set(
                    cert for type_, cert in event_types_certainties
                    if type_ == event_type)
                if "known" in certainties:
                    event_type_certainty = "known"
                elif "suspected" in certainties:
                    event_type_certainty = "suspected"
                else:
                    event_type_certainty = None
            else:
                event_type = None
                event_type_certainty = None
            event.origins.extend(origins)
            event.event_type = event_type
            event.event_type_certainty = event_type_certainty

    def _specify_preferred_origin(self):
        """
        Find the preferred origin

        in two simple cases:
          - only one origin for an event
          - tagged #PRIME in the origin comments
        Affects: event.preferred_origin_id
        """
        event = self.cat[-1]
        event.preferred_origin_id = None
        if len(event.origins) == 1:
            event.preferred_origin_id = event.origins[0].resource_id.id
            return
        # number of origins tagged as #PRIME
        n_prime = 0
        for origin in event.origins:
            for comment in origin.comments:
                if "#PRIME" in comment.text.upper():
                    n_prime += 1
                    event.preferred_origin_id = origin.resource_id.id
                    break  # stop searching comments for this origin
        if n_prime > 1:
            msg = ('Multiple origins tagged as #PRIME for event {}, '
                   'The last one tagged was taken as preferred.'.format(
                       event.resource_id))
            warnings.warn(msg)

    def _read_magnitudes(self):
        event = self.cat[-1]
        while not self._next_line_type():
            line = self._get_next_line()
            if line.strip().startswith('('):
                event.magnitudes[-1].comments.append(
                    self._parse_generic_comment(line))
                continue
            event.magnitudes.append(self._parse_magnitude(line))
            continue

    def _read_bibliography(self):
        event = self.cat[-1]
        while not self._next_line_type():
            line = self._get_next_line()
            if line.strip().startswith('('):
                # TODO parse bibliography comment blocks
                continue
            event.comments.append(self._parse_bibliography_item(line))
            continue

    def _make_comment(self, text):
        id_ = self._construct_id(['comment'], add_hash=True)
        comment = Comment(text=text, resource_id=id_)
        return comment

    def _parse_bibliography_item(self, line):
        return self._make_comment(line)

    def _parse_origin(self, line):
        # 1-10    i4,a1,i2,a1,i2    epicenter date (yyyy/mm/dd)
        # 12-22   i2,a1,i2,a1,f5.2  epicenter time (hh:mm:ss.ss)
        time = UTCDateTime.strptime(line[:17], '%Y/%m/%d %H:%M:')
        time += float(line[17:22])
        # 23      a1    fixed flag (f = fixed origin time solution, blank if
        #                           not a fixed origin time)
        time_fixed = fixed_flag(line[22])
        # 25-29   f5.2  origin time error (seconds; blank if fixed origin time)
        time_error = float_or_none(line[24:29])
        time_error = time_error and QuantityError(uncertainty=time_error)
        # 31-35   f5.2  root mean square of time residuals (seconds)
        rms = float_or_none(line[30:35])
        # 37-44   f8.4  latitude (negative for South)
        latitude = float_or_none(line[36:44])
        # 46-54   f9.4  longitude (negative for West)
        longitude = float_or_none(line[45:54])
        # 55      a1    fixed flag (f = fixed epicenter solution, blank if not
        #                           a fixed epicenter solution)
        epicenter_fixed = fixed_flag(line[54])
        # 56-60   f5.1  semi-major axis of 90% ellipse or its estimate
        #               ([m], blank if fixed epicenter)
        _uncertainty_major_m = float_or_none(line[55:60], multiplier=1e3)
        # 62-66   f5.1  semi-minor axis of 90% ellipse or its estimate
        #               ([m], blank if fixed epicenter)
        _uncertainty_minor_m = float_or_none(line[61:66], multiplier=1e3)
        # 68-70   i3    strike (0 <= x <= 360) of error ellipse clock-wise from
        #                       North (degrees)
        _uncertainty_major_azimuth = float_or_none(line[67:70])
        # 72-76   f5.1  depth [m]
        depth = float_or_none(line[71:76], multiplier=1e3)
        # 77      a1    fixed flag (f = fixed depth station, d = depth phases,
        #                           blank if not a fixed depth)
        depth_type_flag = line[76].lower()
        # 79-82   f4.1  depth error 90% ([m]; blank if fixed depth)
        depth_error = float_or_none(line[78:82], multiplier=1e3)
        # 84-87   i4    number of defining phases
        used_phase_count = int_or_none(line[83:87])
        # 89-92   i4    number of defining stations
        used_station_count = int_or_none(line[88:92])
        # 94-96   i3    gap in azimuth coverage (degrees)
        azimuthal_gap = float_or_none(line[93:96])
        # 98-103  f6.2  distance to closest station (degrees)
        minimum_distance = float_or_none(line[97:103])
        # 105-110 f6.2  distance to furthest station (degrees)
        maximum_distance = float_or_none(line[104:110])
        # 112     a1    analysis type: (a = automatic, m = manual, g = guess)
        evaluation_mode, evaluation_status = \
            evaluation_mode_and_status(line[111])
        # 114     a1    location method: (i = inversion, p = pattern
        #                                 recognition, g = ground truth, o =
        #                                 other)
        location_method = LOCATION_METHODS[line[113].strip().lower()]
        # 116-117 a2    event type:
        # XXX event type and event type certainty is specified per origin,
        # XXX not sure how to bset handle this, for now only use it if
        # XXX information on the individual origins do not clash.. not sure yet
        # XXX how to identify the preferred origin..
        event_type, event_type_certainty = \
            EVENT_TYPE_CERTAINTY[line[115:117].strip().lower()]
        # 119-127 a9    author of the origin
        author = line[118:127].strip()
        # 129-136 a8    origin identification
        origin_id = self._construct_id(['origin', line[128:136].strip()])

        # process items
        # convert GSE to ObsPy depth type
        try:
            depth_type = DEPTH_TYPES[depth_type_flag.strip()]
        except KeyError:
            depth_type = OriginDepthType('from location')
        # assemble depth error
        # if depth_error is not None
        depth_error = depth_error and dict(uncertainty=depth_error,
                                           confidence_level=90)
        # assemble origin uncertainty
        if all(v is not None for v in (_uncertainty_major_m,
                                       _uncertainty_minor_m,
                                       _uncertainty_major_azimuth)):
            origin_uncertainty = OriginUncertainty(
                min_horizontal_uncertainty=_uncertainty_minor_m,
                max_horizontal_uncertainty=_uncertainty_major_m,
                azimuth_max_horizontal_uncertainty=_uncertainty_major_azimuth,
                preferred_description='uncertainty ellipse',
                confidence_level=90)
            # event init always sets an empty QuantityError, even when
            # specifying None, which is strange
            for key in ['confidence_ellipsoid']:
                setattr(origin_uncertainty, key, None)
        else:
            origin_uncertainty = None
        # assemble origin quality
        origin_quality = OriginQuality(
            standard_error=rms, used_phase_count=used_phase_count,
            used_station_count=used_station_count, azimuthal_gap=azimuthal_gap,
            minimum_distance=minimum_distance,
            maximum_distance=maximum_distance)
        # Check for any quality information present
        res = any(vars(origin_quality).values())
        if not res:
            origin_quality = None

        comments = []
        if location_method:
            comments.append(
                self._make_comment('location method: ' + location_method))
        if author:
            creation_info = CreationInfo(author=author)
        else:
            creation_info = None
        # assemble whole event
        origin = Origin(
                   resource_id=origin_id,
                   time=time, time_errors=time_error, time_fixed=time_fixed,
                   longitude=longitude, latitude=latitude,
                   origin_uncertainty=origin_uncertainty,
                   epicenter_fixed=epicenter_fixed,
                   depth=depth, depth_errors=depth_error,
                   depth_type=depth_type, quality=origin_quality,
                   comments=comments, creation_info=creation_info)
        # event init always sets an empty QuantityError, even when specifying
        # None, which is strange
        for key in ('longitude_errors', 'latitude_errors',
                    # 'depth_errors', 'time_errors',
                    ):
            setattr(origin, key, None)
        return origin, event_type, event_type_certainty

    def _parse_magnitude(self, line):
        #    1-5  a5   magnitude type (mb, Ms, ML, mbmle, msmle)
        magnitude_type = line[0:5].strip()
        #      6  a1   min max indicator (<, >, or blank)
        # TODO figure out the meaning of this min max indicator
        min_max_indicator = line[5:6].strip()
        #   7-10  f4.1 magnitude value
        mag = float_or_none(line[6:10])
        #  12-14  f3.1 standard magnitude error
        mag_errors = float_or_none(line[11:14])
        #  16-19  i4   number of stations used to calculate magni-tude
        station_count = int_or_none(line[15:19])
        #  21-29  a9   author of the origin
        author = line[20:29].strip()
        #  31-38  a8   origin identification
        origin_id = line[30:38].strip()

        # process items
        if author:
            creation_info = CreationInfo(author=author)
        else:
            creation_info = None
        mag_errors = mag_errors and QuantityError(uncertainty=mag_errors)
        if origin_id:
            origin_id = self._construct_id(['origin', origin_id])
        else:
            origin_id = None
        if not magnitude_type:
            magnitude_type = None
        # magnitudes have no id field, so construct a unique one at least
        resource_id = self._construct_id(['magnitude'], add_hash=True)

        if min_max_indicator:
            msg = 'Magnitude min/max indicator field not yet implemented'
            warnings.warn(msg)

        # combine and return
        mag = Magnitude(
            magnitude_type=magnitude_type, mag=mag,
            station_count=station_count, creation_info=creation_info,
            mag_errors=mag_errors, origin_id=origin_id,
            resource_id=resource_id)
        # event init always sets an empty QuantityError, even when specifying
        # None, which is strange
        for key in ['mag_errors']:
            setattr(mag, key, None)
        return mag

    def _get_pick_time(self, my_string):
        """
        Look up absolute time of pick including date, based on the time-of-day
        only representation in the phase line

        Returns absolute pick time or None if it can not be determined safely.
        """
        if not my_string.strip():
            return None
        # TODO maybe we should defer phases block parsing.. but that will make
        # the whole reading more complex
        if not self.cat.events:
            msg = ('Can not parse phases block before parsing origins block, '
                   'because phase lines do not contain date information, only '
                   'time-of-day')
            raise NotImplementedError(msg)
        origin_times = [origin.time for origin in self.cat.events[-1].origins]
        if not origin_times:
            msg = ('Can not parse phases block unless origins with origin '
                   'time information are present, because phase lines do not '
                   'contain date information, only time-of-day')
            raise NotImplementedError(msg)
        # XXX this whole routine is on shaky ground..
        # since picks only have a time-of-day and there's not even an
        # association to one of the origins, in principle this would need some
        # real tough logic to make it failsafe. actually this would mean using
        # taup with the given epicentral distance of the pick and check what
        # date is appropriate.
        # for now just do a very simple logic and raise exceptions when things
        # look fishy. this is ugly but it's not worth spending more time on
        # this, unless somebody starts bumping into one of the explicitly
        # raised exceptions below.
        origin_time_min = min(origin_times)
        origin_time_max = max(origin_times)
        hour = int(my_string[0:2])
        minute = int(my_string[3:5])
        seconds = float(my_string[6:])

        all_guesses = []
        for origin in self.cat.events[-1].origins:
            first_guess = UTCDateTime(
                origin.time.year, origin.time.month, origin.time.day, hour,
                minute, seconds)
            all_guesses.append((first_guess, origin.time))
            all_guesses.append((first_guess - 86400, origin.time))
            all_guesses.append((first_guess + 86400, origin.time))

        pick_date = sorted(all_guesses, key=lambda x: abs(x[0] - x[1]))[0][0]

        # make sure event origin times are reasonably close together
        if origin_time_max - origin_time_min > 5 * 3600:
            msg = ('Origin times in event differ by more than 5 hours, this '
                   'is currently not implemented as determining the date of '
                   'the pick might be tricky. Sorry.')
            warnings.warn(msg)
            return None
        # now try the date of the latest origin and raise if things seem fishy
        t = UTCDateTime(pick_date.year, pick_date.month, pick_date.day, hour,
                        minute, seconds)
        for origin_time in origin_times:
            if t - origin_time > 6 * 3600:
                msg = ('This pick would have a time more than 6 hours after '
                       'or before one of the origins in the event. This seems '
                       'fishy. Please report an issue on our github.')
                warnings.warn(msg)
                return None
        return t

    def _parse_phase(self, line, origin_id, values_to_comments=False):
        """
        Parse a phase line, returning Pick, Amplitude, StationMagnitude,
        Arrival objects as available.

        :param line: Line to parse
        :type line: str
        :param origin_id: ResourceIdentifier of the origin
            this phase belongs to
        :type origin_id: obspy.core.event.ResourceIdentifier
        :param values_to_comments: If True, all available origin-specific
            information in the phase line (distance, azimuth, residuals, ...)
            are stored as comments in the Pick.comments attribute.
        :type values_to_comments: bool
        :returns: (pick, amplitude, station_magnitude, arrival)
        :rtype: tuple(obspy.core.event.Pick or None,
                      obspy.core.event.Amplitude or None,
                      obspy.core.event.StationMagnitude or None,
                      obspy.core.event.Arrival or None)
        """

        # It is called with the values_to_comments argument set to True
        # if it was not possible to assign the origin to the phase block
        # and the self.origin_specific_to_comments flag is set.
        # Storing values in comments is not very helpful
        # and somewhat against the spirit of QuakeML.
        # For this, the flag is off by default.
        comments = []

        # 1-5     a5      station code
        station_code = line[0:5].strip()

        # 7-12    f6.2    station-to-event distance (degrees)
        # Arrival.distance (float)
        distance = float_or_none(line[6:12])
        if values_to_comments:
            comments.append(
                    'station-to-event distance (degrees): "{}"'.format(
                        line[6:12]))

        # 14-18   f5.1    event-to-station azimuth (degrees)
        # Arrival.azimuth (float)
        event_azimuth = float_or_none(line[13:18])
        if values_to_comments:
            comments.append(
                    'event-to-station azimuth (degrees): "{}"'.format(
                        line[13:18]))

        # 20-27   a8      phase code
        # Pick.phase_hint (str)
        # Arrival.phase (str)
        phase_hint = line[19:27].strip()

        # 29-40   i2,a1,i2,a1,f6.3        arrival time (hh:mm:ss.sss)
        # Pick.time (UTCDateTime, ATTRIBUTE_HAS_ERRORS)
        time = self._get_pick_time(line[28:40])

        # 42-46   f5.1    time residual (seconds)
        # Arrival.time_residual (float)
        time_residual = float_or_none(line[41:46])
        if values_to_comments:
            comments.append(
                    'time residual (seconds): "{}"'.format(line[41:46]))

        # 48-52   f5.1    observed azimuth (degrees)
        # Pick.backazimuth (float, ATTRIBUTE_HAS_ERRORS)
        backazimuth = float_or_none(line[47:52])

        # 54-58   f5.1    azimuth residual (degrees)
        # Arrival.backazimuth_residual (float)
        backazimuth_residual = float_or_none(line[53:58])
        if values_to_comments:
            comments.append(
                    'azimuth residual (degrees): "{}"'.format(line[53:58]))

        # 60-65   f5.1    observed slowness (seconds/degree)
        # Pick.horizontal_slowness (float, ATTRIBUTE_HAS_ERRORS)
        horizontal_slowness = float_or_none(line[59:65])

        # 67-72   f5.1    slowness residual (seconds/degree)
        # Arrival.horizontal_slowness_residual (float)
        horizontal_slowness_residual = float_or_none(line[66:71])
        if values_to_comments:
            comments.append(
                    'slowness residual (seconds/degree): "{}"'.format(
                        line[66:71]))

        # 74      a1      time defining flag (T or _)
        # Arrival.time_weight (float)
        time_defining_flag = line[73]
        time_weight = None
        if time_defining_flag == 'T':
            time_weight = 1

        # 75      a1      azimuth defining flag (A or _)
        # Arrival.backazimuth_weight (float)
        azimuth_defining_flag = line[74]
        backazimuth_weight = None
        if azimuth_defining_flag == 'A':
            backazimuth_weight = 1

        # 76      a1      slowness defining flag (S or _)
        # Arrival.horizontal_slowness_weight (float)
        slowness_defining_flag = line[75]
        horizontal_slowness_weight = None
        if slowness_defining_flag == 'S':
            horizontal_slowness_weight = 1
        # TAS flag (T, A, S, or _) not implemented in QuakeML
        # but reflects in the weights above
        if values_to_comments:
            comments.append('TAS flag: "{}"'.format(line[73:76]))

        # 78-82   f5.1    signal-to-noise ratio
        # Amplitude.snr (float)
        snr = float_or_none(line[77:82])

        # 84-92   f9.1    amplitude (nanometers)
        # Amplitude.generic_amplitude (float, ATTRIBUTE_HAS_ERRORS)
        amp = float_or_none(line[83:92])

        # 94-98   f5.2    period (seconds)
        # Amplitude.period (float, ATTRIBUTE_HAS_ERRORS)
        period = float_or_none(line[93:98])

        # 100     a1      type of pick (a = automatic, m = manual)
        # Pick.evaluation_mode (EvaluationMode)
        evaluation_mode = line[99]

        # 101     a1      direction of short period motion
        #                 (c = compression, d = dilatation, _= null)
        # Pick.polarity (PickPolarity)
        polarity = POLARITY[line[100].strip().lower()]

        # 102     a1      onset quality (i = impulsive, e = emergent,
        #                                q = questionable, _ = null)
        # Pick.onset (PickOnset)
        onset = ONSET[line[101].strip().lower()]

        # 104-108 a5      magnitude type (mb, Ms, ML, mbmle, msmle)
        # StationMagnitude.station_magnitude_type (str)
        magnitude_type = line[103:108].strip()

        # 109     a1      min max indicator (<, >, or blank)
        min_max_indicator = line[108].strip()

        # 110-113 f4.1    magnitude value
        # StationMagnitude.mag (float, ATTRIBUTE_HAS_ERRORS)
        mag = float_or_none(line[109:113])

        # 115-122 a8      arrival identification
        phase_id = line[114:122].strip()
        #
        if (time is None) and (amp is None) and (mag is None):
            msg = ('Could not determine absolute time of pick. This phase '
                   'line will be ignored:\n{}').format(line)
            warnings.warn(msg)
            return None, None, None, None
        # process items
        waveform_id = WaveformStreamID(network_code=None,
                                       station_code=station_code)
        evaluation_mode = PICK_EVALUATION_MODE[evaluation_mode.strip().lower()]
        if len(comments) > 0:
            comments = [self._make_comment(', '.join(comments))]
        # assemble
        # pick
        if phase_id:
            pick_id = self._construct_id(['pick', phase_id])
        else:
            pick_id = self._construct_id(['pick'], add_hash=True)
        pick = Pick(resource_id=pick_id,
                    phase_hint=phase_hint, time=time, waveform_id=waveform_id,
                    evaluation_mode=evaluation_mode, comments=comments,
                    polarity=polarity, onset=onset, backazimuth=backazimuth,
                    horizontal_slowness=horizontal_slowness
                    )
        # event init always sets an empty QuantityError, even when specifying
        # None, which is strange
        for key in ('time_errors', 'horizontal_slowness_errors',
                    'backazimuth_errors'):
            setattr(pick, key, None)
        # amplitude
        # When reading from QuakeML later on,
        # Amplitude.waveform_id.network_code defaults
        # to empty string if not given.
        waveform_id = WaveformStreamID(network_code="",
                                       station_code=station_code)
        if amp:
            if phase_id:
                amplitude_id = self._construct_id(['amplitude', phase_id])
            else:
                amplitude_id = self._construct_id(['amplitude'], add_hash=True)
            # Convert from nanometers to meters
            amp /= 1e9
            amplitude = Amplitude(resource_id=amplitude_id,
                                  unit=AmplitudeUnit.M, generic_amplitude=amp,
                                  period=period, snr=snr,
                                  pick_id=pick_id, waveform_id=waveform_id,
                                  )
        else:
            amplitude = None
            amplitude_id = None
        # station_magnitude
        waveform_id = WaveformStreamID(network_code=None,
                                       station_code=station_code)
        if mag:
            comment = ''
            if min_max_indicator:
                # StationMagnitude min/max indicator not implemented in QuakeML
                # min max indicator (<, >, or '')
                comment = ('min max indicator: ' + min_max_indicator)
            if phase_id:
                station_magnitude_id = self._construct_id(
                        ['station_magnitude', phase_id])
            else:
                station_magnitude_id = self._construct_id(
                        ['station_magnitude'], add_hash=True)
            station_magnitude = StationMagnitude(
                    resource_id=station_magnitude_id,
                    origin_id=origin_id,
                    mag=mag, magnitude_type=magnitude_type,
                    amplitude_id=amplitude_id, waveform_id=waveform_id,
                    comments=[self._make_comment(comment)] if comment else []
                    )
            # QuakeML in the ObsPy implementation stores <stationMagnitude>
            # <originID> as a string, even if the StationMagnitude.origin_id
            # attribute is set to None or missing it results
            # in <originID>None</originID> in the XML file.
            # if origin_id is None:
            #     delattr(station_magnitude, 'origin_id')
            for key in ['mag_errors']:
                setattr(station_magnitude, key, None)
        else:
            station_magnitude = None
        # arrival
        if distance or time_residual:
            if phase_id:
                arrival_id = self._construct_id(['arrival', phase_id])
            else:
                arrival_id = self._construct_id(['arrival'], add_hash=True)
            arrival = Arrival(
                    resource_id=arrival_id, pick_id=pick_id, phase=phase_hint,
                    azimuth=event_azimuth, distance=distance,
                    time_residual=time_residual,
                    backazimuth_residual=backazimuth_residual,
                    horizontal_slowness_residual=horizontal_slowness_residual,
                    time_weight=time_weight,
                    backazimuth_weight=backazimuth_weight,
                    horizontal_slowness_weight=horizontal_slowness_weight
                    )
        else:
            arrival = None

        return pick, amplitude, station_magnitude, arrival

    def _parse_generic_comment(self, line):
        return self._make_comment(line)


def _buffer_proxy(filename_or_buf, function, reset_fp=True,
                  file_mode="rb", *args, **kwargs):
    """
    Calls a function with an open file or file-like object as the first
    argument. If the file originally was a filename, the file will be
    opened, otherwise it will just be passed to the underlying function.

    :param filename_or_buf: File to pass.
    :type filename_or_buf: str, open file, or file-like object
    :param function: The function to call.
    :param reset_fp: If True, the file pointer will be set to the initial
        position after the function has been called.
    :type reset_fp: bool
    :param file_mode: Mode to open file in if necessary.
    """
    try:
        position = filename_or_buf.tell()
        is_buffer = True
    except AttributeError:
        is_buffer = False

    if is_buffer is True:
        ret_val = function(filename_or_buf, *args, **kwargs)
        if reset_fp:
            filename_or_buf.seek(position, 0)
        return ret_val
    else:
        with open(filename_or_buf, file_mode) as fh:
            return function(fh, *args, **kwargs)


def _read_ims10_bulletin(filename_or_buf, **kwargs):
    """
    Reads an ISF IMS1.0 bulletin file to a :class:`~obspy.core.event.Catalog`
    object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.event.read_events` function, call this
        instead.

    .. rubric:: Notes

    (1) This reader only supports the IMS1.0:SHORT subformat, not LONG or
        other subformats.

    (2) Compared to previous versions (1.4.x) of this reader, the processing
        of phase blocks has changed. Phase blocks contain origin-specific data
        (distance, azimuth, residuals, defining flags).
        In previous versions, origin-specific data was always stored as text
        in the Pick.comments object.
        Now this data is stored in the Arrival object according to QukeML.
        If it was not possible to assign the preferred origin (e.g., because
        multiple origins are given and the #PRIME or #OrigID tag is not present
        ), a warning is issued and the parsing of this phase block is skipped.

    :type file: str or :class:`~pathlib.Path` or
        file-like object( e.g. `ByesIO`)
    :param filename_or_buf: bulletin file to read
    :raises: :class:`~obspy.core.ObsPyReadingError` on problems to read file
    :rtype: :class:`~obspy.core.event.Catalog`
    :return: Catalog object
    """

    # The default behavior is to skip orphaned phase blocks.
    # This can be changed by calling the _read_ims10_bulletin function with
    # two keyword arguments `skip_orphan` and `origin_specific_to_comments`.
    # By setting `skip_orphan` to False and
    # `origin_specific_to_comments` to True, the phase block is analyzed even
    # if the preferred origin cannot be assigned.
    # However, since origin-specific data cannot be stored in the Arrival
    # object without a valid origin_id, this data is stored as text
    # in the Pick.comments attributes, similar to previous versions.

    try:
        return _buffer_proxy(filename_or_buf, __read_ims10_bulletin,
                             reset_fp=False, **kwargs)
    # Happens for example when passing the data as a string which would be
    # interpreted as a filename.
    except OSError:
        return False


def __read_ims10_bulletin(fh, **kwargs):  # NOQA
    """
    Reads an ISF IMS1.0 bulletin file to a :class:`~obspy.core.event.Catalog`
    object.

    :param fh: File or file-like object
    """
    return ISFReader(fh, **kwargs).deserialize()


def _is_ims10_bulletin(filename_or_buf, **kwargs):
    """
    Checks whether a file is ISF IMS1.0 bulletin format.

    :type filename_or_buf: str or file
    :param filename_or_buf: name of the file to be checked or open file-like
        object.
    :rtype: bool
    :return: ``True`` if ISF IMS1.0 bulletin file.
    """
    try:
        return _buffer_proxy(filename_or_buf, __is_ims10_bulletin,
                             reset_fp=True, **kwargs)
    # Happens for example when passing the data as a string which would be
    # interpreted as a filename.
    except OSError:
        return False


def __is_ims10_bulletin(fh, **kwargs):  # NOQA
    """
    Checks whether a file is ISF IMS1.0 bulletin format.

    :type fh: open file or file-like object
    :param filename: name of the file to be checked or open file-like object.
    :rtype: bool
    :return: ``True`` if ISF IMS1.0 bulletin file.
    """
    iline = 0
    try:
        for line in fh:
            iline += 1
            # the header can be preceded by a max of 40 lines of text
            if iline > 40:
                break
            sline = _decode_if_possible(line).rstrip()
            if sline.upper().startswith("DATA_TYPE BULLETIN IMS1.0"):
                if 'LONG' in sline.upper():
                    return False
                else:
                    # if subformat is not specified, assume SHORT
                    return True
    except Exception:
        return False
    return False



# -*- coding: utf-8 -*-
"""
ObsPy's compatibility layer.

Includes things to easy dealing with Python version differences as well as
making it work with various versions of our dependencies.
"""
import numpy as np


# NumPy does not offer the from_buffer method under Python 3 and instead
# relies on the built-in memoryview object.
def from_buffer(data, dtype):
    try:
        data = data.encode()
    except Exception:
        pass
    return np.array(memoryview(data)).view(dtype).copy()  # NOQA


def round_away(number):
    """
    Simple function that rounds a number to the nearest integer. If the number
    is halfway between two integers, it will round away from zero. Of course
    only works up machine precision. This should hopefully behave like the
    round() function in Python 2.

    This is potentially desired behavior in the trim functions but some more
    thought should be poured into it.

    The np.round() function rounds towards the even nearest even number in case
    of half-way splits.

    >>> round_away(2.5)
    3
    >>> round_away(-2.5)
    -3

    >>> round_away(10.5)
    11
    >>> round_away(-10.5)
    -11

    >>> round_away(11.0)
    11
    >>> round_away(-11.0)
    -11
    """
    floor = np.floor(number)
    ceil = np.ceil(number)
    if (floor != ceil) and (abs(number - floor) == abs(ceil - number)):
        return int(int(number) + int(np.sign(number)))
    else:
        return int(np.round(number))



import matplotlib.pyplot as plt

import obspy
from obspy.signal.trigger import plot_trigger, modified_energy_ratio


trace = obspy.read("https://examples.obspy.org/ev0_6.a01.gse2")[0]
df = trace.stats.sampling_rate

cft = modified_energy_ratio(trace.data, int(4 * df))
fig, axes = plot_trigger(trace, cft, cft.max(), 0.1, show=False)
axes[1].set_yscale('log')
axes[1].set_ylim(ymin=0.01)
plt.show()



# -*- coding: utf-8 -*-
"""
CSS bindings to ObsPy core module.
"""
from pathlib import Path
import numpy as np

from obspy import Stream, Trace, UTCDateTime
from obspy.core.compatibility import from_buffer

import gzip

DTYPE = {
    # Big-endian integers
    b's4': b'>i',
    b's2': b'>h',
    # Little-endian integers
    b'i4': b'<i',
    b'i2': b'<h',
    # ASCII integers
    b'c0': (b'S12', int),
    b'c#': (b'S12', int),
    # Big-endian floating point
    b't4': b'>f',
    b't8': b'>d',
    # Little-endian floating point
    b'f4': b'<f',
    b'f8': b'<d',
    # ASCII floating point
    b'a0': (b'S15', np.float32),
    b'a#': (b'S15', np.float32),
    b'b0': (b'S24', np.float64),
    b'b#': (b'S24', np.float64),
}


def _is_css(filename):
    """
    Checks whether a file is CSS waveform data (header) or not.

    :type filename: str
    :param filename: CSS file to be checked.
    :rtype: bool
    :return: ``True`` if a CSS waveform header file.
    """
    # Fixed file format.
    # Tests:
    #  - the length of each line (283 chars)
    #  - two epochal time fields
    #    (for position of dot and if they convert to UTCDateTime)
    #  - supported data type descriptor
    try:
        with open(filename, "rb") as fh:
            lines = fh.readlines()
            # check for empty file
            if not lines:
                return False
            # check every line
            for line in lines:
                assert len(line.rstrip(b"\n\r")) == 283
                assert b"." in line[26:28]
                UTCDateTime(float(line[16:33]))
                assert b"." in line[71:73]
                UTCDateTime(float(line[61:78]))
                assert line[143:145] in DTYPE
    except Exception:
        return False
    return True


def _is_nnsa_kb_core(filename):
    """
    Checks whether a file is NNSA KB Core waveform data (header) or not.

    :type filename: str
    :param filename: NNSA KB Core file to be checked.
    :rtype: bool
    :return: ``True`` if a NNSA KB Core waveform header file.
    """
    # Fixed file format.
    # Tests:
    #  - the length of each line (287 chars)
    #  - two epochal time fields
    #    (for position of dot and if they convert to UTCDateTime)
    #  - supported data type descriptor
    try:
        with open(filename, "rb") as fh:
            lines = fh.readlines()
            # check for empty file
            if not lines:
                return False
            # check every line
            for line in lines:
                assert len(line.rstrip(b"\n\r")) == 287
                assert line[27:28] == b"."
                UTCDateTime(float(line[16:33]))
                assert line[73:74] == b"."
                UTCDateTime(float(line[62:79]))
                assert line[144:146] in DTYPE
    except Exception:
        return False
    return True


def _read_css(filename, **kwargs):
    """
    Reads a CSS waveform file and returns a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type filename: str
    :param filename: CSS file to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream with Traces specified by given file.
    """
    # read metafile with info on single traces
    with open(filename, "rb") as fh:
        lines = fh.readlines()
    basedir = Path(filename).parent
    traces = []
    # read single traces
    for line in lines:
        npts = int(line[79:87])
        dirname = line[148:212].strip().decode()
        dfilename = Path(basedir) / dirname / line[213:245].strip().decode()
        offset = int(line[246:256])
        dtype = DTYPE[line[143:145]]
        if isinstance(dtype, tuple):
            read_fmt = np.dtype(dtype[0])
            fmt = dtype[1]
        else:
            read_fmt = np.dtype(dtype)
            fmt = read_fmt

        try:
            # assumed that the waveform file is not compressed
            fh = open(dfilename, "rb")
        except FileNotFoundError as e:
            # If does not find the waveform file referenced in the wfdisc,
            # it will try to open a compressed .gz suffix file instead.
            try:
                fh = gzip.open(str(dfilename) + '.gz', "rb")
            except FileNotFoundError:
                raise e

        # Read one segment of binary data
        fh.seek(offset)
        data = fh.read(read_fmt.itemsize * npts)
        fh.close()
        data = from_buffer(data, dtype=read_fmt)
        data = np.require(data, dtype=fmt)

        header = {}
        header['station'] = line[0:6].strip().decode()
        header['channel'] = line[7:15].strip().decode()
        header['starttime'] = UTCDateTime(float(line[16:33]))
        header['sampling_rate'] = float(line[88:99])
        header['calib'] = float(line[100:116])
        header['calper'] = float(line[117:133])
        tr = Trace(data, header=header)
        traces.append(tr)
    return Stream(traces=traces)


def _read_nnsa_kb_core(filename, **kwargs):
    """
    Reads a NNSA KB Core waveform file and returns a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type filename: str
    :param filename: NNSA KB Core file to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream with Traces specified by given file.
    """
    # read metafile with info on single traces
    with open(filename, "rb") as fh:
        lines = fh.readlines()
    basedir = Path(filename).parent
    traces = []
    # read single traces
    for line in lines:
        npts = int(line[80:88])
        dirname = line[149:213].strip().decode()
        filename = Path(basedir) / dirname / \
            line[214:246].strip().decode()

        offset = int(line[247:257])
        dtype = DTYPE[line[144:146]]
        if isinstance(dtype, tuple):
            read_fmt = np.dtype(dtype[0])
            fmt = dtype[1]
        else:
            read_fmt = np.dtype(dtype)
            fmt = read_fmt
        with open(filename, "rb") as fh:
            fh.seek(offset)
            data = fh.read(read_fmt.itemsize * npts)
            data = from_buffer(data, dtype=read_fmt)
            data = np.require(data, dtype=fmt)
        header = {}
        header['station'] = line[0:6].strip().decode()
        header['channel'] = line[7:15].strip().decode()
        header['starttime'] = UTCDateTime(float(line[16:33]))
        header['sampling_rate'] = float(line[89:100])
        header['calib'] = float(line[101:117])
        header['calper'] = float(line[118:134])
        tr = Trace(data, header=header)
        traces.append(tr)
    return Stream(traces=traces)



# -*- coding: utf-8 -*-
"""
The obspy.imaging.scripts.scan / obspy-scan test suite.
"""
import os
import shutil
import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

from obspy import read, UTCDateTime
from obspy.core.util.base import NamedTemporaryFile, get_example_file
from obspy.core.util.misc import TemporaryWorkingDirectory, CatchOutput
from obspy.imaging.scripts.scan import main as obspy_scan
from obspy.imaging.scripts.scan import scan, Scanner


class TestScan:
    """
    Test cases for obspy-scan
    """
    @pytest.fixture(scope='class')
    def all_files(self, root):
        """return a list of all waveform files in the test suite."""
        sac_files = ['LMOW.BHE.SAC', 'seism.sac', 'dis.G.SCZ.__.BHE_short',
                     'null_terminated.sac', 'test.sac', 'seism-longer.sac',
                     'test.sac.swap', 'seism-shorter.sac', 'testxy.sac']
        gse2_files = ['STA2.testlines', 'STA2.testlines_out', 'acc.gse',
                      'loc_RJOB20050831023349.z',
                      'loc_RJOB20050831023349_first100_dos.z',
                      'loc_RNON20040609200559.z', 'loc_STAU20031119011659.z',
                      'sta2.gse2', 'twiceCHK2.gse2', 'y2000.gse']
        all_files = [root / 'io' / 'sac' / 'tests' / 'data' / i
                     for i in sac_files]
        all_files.extend([root / 'io' / 'gse2' / 'tests' / 'data' / i
                          for i in gse2_files])
        return all_files

    def test_scan_main_method(self, all_files, image_path):
        """
        Run obspy-scan on selected tests/data directories
        """
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            for filename in all_files:
                shutil.copy(filename, os.curdir)

            obspy_scan([os.curdir] + ['--output', str(image_path)])

    def test_scan_dir_no_permission(self, all_files):
        """
        Run obspy-scan on a directory without read permission.
        Should just skip it and not raise an exception. see #3115
        """
        # Copy files to a temp folder to avoid wildcard scans.
        scanner = Scanner()
        with TemporaryWorkingDirectory():
            no_permission_dir = Path('no_permission_dir')
            no_permission_dir.mkdir()
            # copy one test file in
            shutil.copy(all_files[0], no_permission_dir)
            # take away read permissions
            no_permission_dir.chmod(0o000)
            # still sometimes even on Linux CI runners it seems that stripping
            # read permission does not work and the scanner is able to process
            # it, thus ending up with a higher counter than if properly making
            # the directory not readable
            try:
                os.listdir(no_permission_dir)
            # this is what we want, not being able to read the directory, thus
            # the scanner only processing one item
            except PermissionError:
                pass
            # this is unexpected, we weren't able to strip read permissions,
            # and then in consequence this test is meaningless and we mark it
            # as a skipped test
            else:
                pytest.skip(
                    "unable to remove read permission from a test file for "
                    "testing purposes")
            scanner.parse(str(no_permission_dir))
            # should not have been able to read test file but also not raised
            # an error
            assert scanner.counter == 1
            assert not scanner.data
            # now allow read permission and read again
            no_permission_dir.chmod(0o777)
            scanner.parse(str(no_permission_dir))
            assert scanner.counter == 2
            assert '.LMOW..BHE' in scanner.data
            for child in no_permission_dir.iterdir():
                child.unlink()
            no_permission_dir.rmdir()

    def test_scan_function_and_scanner_class(self, all_files, image_path):
        """
        Test scan function and Scanner class (in one test to keep overhead of
        copying files down)
        """
        scanner = Scanner()
        path1 = image_path.parent / 'scan1.png'
        path2 = image_path.parent / 'scan2.png'
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            for filename in all_files:
                shutil.copy(filename, os.curdir)

            scanner.parse(os.curdir)
            scan(paths=os.curdir, plot=path1)

        scanner.plot(path2)

    def test_scan_plot_by_id_with_wildcard(self):
        """
        Test selecting what to plot after scanning with wildcards in selected
        SEED IDs
        """
        files = [
            "BW.UH1._.EHZ.D.2010.147.a.slist.gz",
            "BW.UH1._.EHZ.D.2010.147.b.slist.gz",
            "BW.UH1._.SHZ.D.2010.147.cut.slist.gz",
            "BW.UH2._.SHZ.D.2010.147.cut.slist.gz",
            "BW.UH3._.SHE.D.2010.147.cut.slist.gz",
            "BW.UH3._.SHN.D.2010.147.cut.slist.gz",
            "BW.UH3._.SHZ.D.2010.147.cut.slist.gz",
            "BW.UH4._.EHZ.D.2010.147.cut.slist.gz",
            "IUANMO.seed"]

        scanner = Scanner()

        for filename in files:
            scanner.parse(get_example_file(filename))

        expected = [
            ('*.UH[12]*',
             ['BW.UH2..SHZ\n100.0%',
              'BW.UH1..SHZ\n100.0%',
              'BW.UH1..EHZ\n10.7%']),
            ('*Z',
             ['IU.ANMO.00.LHZ\n100.0%',
              'BW.UH4..EHZ\n100.0%',
              'BW.UH3..SHZ\n100.0%',
              'BW.UH2..SHZ\n100.0%',
              'BW.UH1..SHZ\n100.0%',
              'BW.UH1..EHZ\n10.7%'])]

        for seed_id, expected_labels in expected:
            fig, ax = plt.subplots()
            fig = scanner.plot(fig=fig, show=False, seed_ids=[seed_id])
            got = [label.get_text() for label in ax.get_yticklabels()]
            assert got == expected_labels
            plt.close(fig)

    def test_scanner_manually_add_streams(self, all_files, image_path):
        """
        Test Scanner class, manually adding streams of read data files
        """
        scanner = Scanner()
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            for filename in all_files:
                shutil.copy(filename, os.curdir)

            for file_ in os.listdir(os.curdir):
                # some files used in the test cases actually can not
                # be read with obspy..
                if file_ in ('STA2.testlines_out', 'STA2.testlines',
                             'seism-shorter.sac', 'seism-longer.sac'):
                    continue
                st = read(file_, headonly=True)
                scanner.add_stream(st)

        scanner.plot(str(image_path))

    def test_scan_save_load_npz(self, all_files, image_path):
        """
        Run obspy-scan on selected tests/data directories, saving/loading
        to/from npz.

        Tests both the command line script and the Scanner class.
        """
        scanner = Scanner()
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            for filename in all_files:
                shutil.copy(filename, os.curdir)

            # save via command line
            obspy_scan([os.curdir, '--write', 'scan.npz'])

            # save via Python
            scanner.parse(os.curdir)
            scanner.save_npz('scanner.npz')
            scanner = Scanner()

            path1 = image_path.parent / 'scan_load_npz_1.png'
            path2 = image_path.parent / 'scan_load_npz_2.png'

            # version string of '0.0.0+archive' raises UserWarning - ignore
            with warnings.catch_warnings(record=True):
                warnings.simplefilter('ignore', UserWarning)

                # load via Python
                scanner.load_npz('scanner.npz')
                scanner.plot(path1)

                # load via command line
                obspy_scan(['--load', 'scan.npz', '--output', str(path2)])

    def test_scan_times(self, all_files, image_path):
        """
        Checks for timing related options
        """
        # Copy files to a temp folder to avoid wildcard scans.
        with TemporaryWorkingDirectory():
            for filename in all_files:
                shutil.copy(filename, os.curdir)

            obspy_scan([os.curdir] + ['--output', str(image_path)] +
                       ['--start-time', '2004-01-01'] +
                       ['--end-time', '2004-12-31'] +
                       ['--event-time', '2004-03-14T15:09:26'] +
                       ['--event-time', '2004-02-07T18:28:18'])

    def test_multiple_sampling_rates(self, image_path):
        """
        Check for multiple sampling rates
        """
        lines = [
            "TIMESERIES XX_TEST__BHZ_R, 200 samples, 200 sps, "
            "2008-01-15T00:00:00.000000, SLIST, INTEGER, Counts",
            "TIMESERIES XX_TEST__BHZ_R,  50 samples,  50 sps, "
            "2008-01-15T00:00:00.900000, SLIST, INTEGER, Counts",
            "TIMESERIES XX_TEST__BHZ_R, 200 samples, 200 sps, "
            "2008-01-15T00:00:02.000000, SLIST, INTEGER, Counts",
        ]
        files = []
        expected = [
            "XX.TEST..BHZ 2008-01-15T00:00:01.000000Z "
            "2008-01-15T00:00:00.899995Z -0.100",
            "XX.TEST..BHZ 2008-01-15T00:00:01.899999Z "
            "2008-01-15T00:00:02.000000Z 0.100"
        ]
        with NamedTemporaryFile() as f1, NamedTemporaryFile() as f2, \
                NamedTemporaryFile() as f3:
            for i, fp in enumerate([f1, f2, f3]):
                fp.write(("%s\n" % lines[i]).encode('ascii',
                                                    'strict'))
                fp.flush()
                fp.seek(0)
                files.append(fp.name)

            # make image comparison instance and set manual rms (see #2089)

            with CatchOutput() as out:
                cmds = ['--output', str(image_path), '--print-gaps']
                obspy_scan(files + cmds)

            # read output and compare with expected
            # only check if datetime objects are close, not exact
            output = out.stdout.splitlines()
            for ex_line, out_line in zip(expected, output):
                ex_split = ex_line.split(' ')
                out_split = out_line.split(' ')
                for ex_str, out_str in zip(ex_split, out_split):
                    try:
                        utc1 = UTCDateTime(ex_str)
                        utc2 = UTCDateTime(out_str)
                    except (ValueError, TypeError):
                        # if str is not a datetime it should be equal
                        assert ex_str == out_str
                    else:
                        # datetimes just need to be close
                        t1, t2 = utc1.timestamp, utc2.timestamp
                        assert abs(t1 - t2) < .001



import numpy as np
import matplotlib.pyplot as plt

import obspy


# Read the seismogram
st = obspy.read("https://examples.obspy.org/RJOB_061005_072159.ehz.new")

# There is only one trace in the Stream object, let's work on that trace...
tr = st[0]

# Filtering with a lowpass on a copy of the original Trace
tr_filt = tr.copy()
tr_filt.filter('lowpass', freq=1.0, corners=2, zerophase=True)

# Now let's plot the raw and filtered data...
t = np.arange(0, tr.stats.npts / tr.stats.sampling_rate, tr.stats.delta)
plt.subplot(211)
plt.plot(t, tr.data, 'k')
plt.ylabel('Raw Data')
plt.subplot(212)
plt.plot(t, tr_filt.data, 'k')
plt.ylabel('Lowpassed Data')
plt.xlabel('Time [s]')
plt.suptitle(tr.stats.starttime)
plt.show()



from __future__ import print_function

from math import log10

from obspy import read


st = read("../data/LKBD_WA_CUT.MSEED")

tr_n = st.select(component="N")[0]
ampl_n = max(abs(tr_n.data))

tr_e = st.select(component="E")[0]
ampl_e = max(abs(tr_e.data))

ampl = max(ampl_n, ampl_e)

epi_dist = 20

a = 0.018
b = 2.17
ml = log10(ampl * 1000) + a * epi_dist + b
print(ml)



import numpy as np
import matplotlib.pyplot as plt

from obspy.taup import TauPyModel


PHASES = [
    # Phase, distance
    ('P', 26),
    ('PP', 60),
    ('PPP', 94),
    ('PPS', 155),
    ('p', 3),
    ('pPcP', 100),
    ('PKIKP', 170),
    ('PKJKP', 194),
    ('S', 65),
    ('SP', 85),
    ('SS', 134.5),
    ('SSS', 204),
    ('p', -10),
    ('pP', -37.5),
    ('s', -3),
    ('sP', -49),
    ('ScS', -44),
    ('SKS', -82),
    ('SKKS', -120),
]

model = TauPyModel(model='iasp91')

fig, ax = plt.subplots(subplot_kw=dict(polar=True))

# Plot all pre-determined phases
for phase, distance in PHASES:
    arrivals = model.get_ray_paths(700, distance, phase_list=[phase])
    ax = arrivals.plot_rays(plot_type='spherical',
                            legend=False, label_arrivals=True,
                            plot_all=True,
                            show=False, ax=ax, indicate_wave_type=True)

# Annotate regions
ax.text(0, 0, 'Solid\ninner\ncore',
        horizontalalignment='center', verticalalignment='center',
        bbox=dict(facecolor='white', edgecolor='none', alpha=0.7))
ocr = (model.model.radius_of_planet -
       (model.model.s_mod.v_mod.iocb_depth +
        model.model.s_mod.v_mod.cmb_depth) / 2)
ax.text(np.deg2rad(180), ocr, 'Fluid outer core',
        horizontalalignment='center',
        bbox=dict(facecolor='white', edgecolor='none', alpha=0.7))
mr = model.model.radius_of_planet - model.model.s_mod.v_mod.cmb_depth / 2
ax.text(np.deg2rad(180), mr, 'Solid mantle',
        horizontalalignment='center',
        bbox=dict(facecolor='white', edgecolor='none', alpha=0.7))

plt.show()



# -*- coding: utf-8 -*-
"""
SQLAlchemy ORM definitions (database layout) for tsindex.db.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import datetime

from sqlalchemy import Column, String, Integer, Float
from sqlalchemy.orm import declarative_base
from sqlalchemy.schema import PrimaryKeyConstraint

Base = declarative_base()

# Every declarative class should only be instantiated once. Thus we just use a
# simple cache to be able to cache by table name.
_table_cache = {"index_tables": {}, "summary_tables": {}}


def _get_tsindex_table(table_name='tsindex'):

    if table_name not in _table_cache["index_tables"]:
        class TSIndexTable(Base):
            """
            DB table containing tsindex.
            """
            __tablename__ = table_name
            __table_args__ = (
                PrimaryKeyConstraint('network', 'station',
                                     'location', 'channel',
                                     'quality', 'version',
                                     'starttime', 'endtime'),
                {'keep_existing': True}
            )

            network = Column(String)
            station = Column(String)
            location = Column(String)
            channel = Column(String)
            quality = Column(String(1))
            version = Column(Integer)
            starttime = Column(String)
            endtime = Column(String)
            samplerate = Column(Float)
            filename = Column(String)
            byteoffset = Column(Integer)
            bytes = Column(Integer)
            hash = Column(String)
            timeindex = Column(String)
            timespans = Column(String)
            timerates = Column(String)
            format = Column(String)
            filemodtime = Column(String)
            updated = Column(String)
            scanned = Column(String)

            def __repr__(self):
                return "<TSIndexTable('%s %s %s %s %s %s %s')>" % \
                        (self.network, self.station, self.location,
                         self.channel, self.starttime, self.endtime,
                         self.filemodtime)

        _table_cache["index_tables"][table_name] = TSIndexTable
    return _table_cache["index_tables"][table_name]


def _get_tsindex_summary_table(table_name='tsindex_summary'):
    if table_name not in _table_cache["summary_tables"]:
        class TSIndexSummaryTable(Base):
            """
            DB table containing tsindex.
            """
            __tablename__ = table_name
            __table_args__ = (
                PrimaryKeyConstraint('network', 'station',
                                     'location', 'channel',
                                     'earliest', 'latest'),
                {'keep_existing': True}
            )

            network = Column(String)
            station = Column(String)
            location = Column(String)
            channel = Column(String)
            earliest = Column(String)
            latest = Column(String)
            updt = Column(String,
                          default=datetime.datetime.utcnow)

            def __repr__(self):
                return "<TSIndexSummaryTable('%s %s %s %s %s %s %s')>" % \
                        (self.network, self.station, self.location,
                         self.channel, self.earliest, self.latest, self.updt)

        _table_cache["summary_tables"][table_name] = TSIndexSummaryTable
    return _table_cache["summary_tables"][table_name]


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility objects.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), Tom Eulenfeld 2013-2024
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import copy
import re
import warnings
from textwrap import TextWrapper

from obspy import UTCDateTime
from obspy.core.util.base import ComparingObject
from obspy.core.util.decorator import deprecated_keywords
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning
from obspy.core.util.obspy_types import (FloatWithUncertaintiesAndUnit,
                                         FloatWithUncertaintiesFixedUnit)


class BaseNode(ComparingObject):
    """
    From the StationXML definition:
        A base node type for derivation of: Network, Station and Channel
        types.

    The parent class for the network, station and channel classes.
    """
    def __init__(self, code, description=None, comments=None, start_date=None,
                 end_date=None, restricted_status=None, alternate_code=None,
                 historical_code=None, data_availability=None,
                 identifiers=None, source_id=None):
        """
        :type code: str
        :param code: The SEED network, station, or channel code
        :type description: str, optional
        :param description: A description of the resource
        :type comments: list of :class:`Comment`, optional
        :param comments: An arbitrary number of comments to the resource
        :type start_date: :class:`~obspy.core.utcdatetime.UTCDateTime`,
            optional
        :param start_date: The start date of the resource
        :type end_date: :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        :param end_date: The end date of the resource
        :type restricted_status: str, optional
        :param restricted_status: The restriction status
        :type alternate_code: str, optional
        :param alternate_code: A code used for display or association,
            alternate to the SEED-compliant code.
        :type historical_code: str, optional
        :param historical_code: A previously used code if different from the
            current code.
        :type data_availability:
            :class:`~obspy.core.inventory.util.DataAvailability`
        :param data_availability: Information about time series availability
            for the network/station/channel.
        :type identifiers: list[str], optional
        :param identifiers: Persistent identifiers for network/station/channel
            (schema version >=1.1). URIs are in general composed of a 'scheme'
            and a 'path' (optionally with additional components), the two of
            which separated by a colon.
        :type source_id: str, optional
        :param source_id: A data source identifier in URI form
            (schema version >=1.1). URIs are in general composed of a 'scheme'
            and a 'path' (optionally with additional components), the two of
            which separated by a colon.
        """
        self.code = code
        self.comments = comments or []
        self.description = description
        self.start_date = start_date
        self.end_date = end_date
        self.restricted_status = restricted_status
        self.alternate_code = alternate_code
        self.historical_code = historical_code
        self.data_availability = data_availability
        self.identifiers = identifiers or []
        self.source_id = source_id

    @property
    def code(self):
        return self._code

    @code.setter
    def code(self, value):
        if value is None:
            msg = "A code is required"
            raise ValueError(msg)
        self._code = str(value).strip()

    @property
    def source_id(self):
        return self._source_id

    @source_id.setter
    def source_id(self, value):
        if value:
            _warn_on_invalid_uri(value)
            self._source_id = value.strip()
        else:
            self._source_id = None

    @property
    def identifiers(self):
        return self._identifiers

    @identifiers.setter
    def identifiers(self, value):
        if not hasattr(value, "__iter__"):
            msg = "identifiers needs to be an iterable, e.g. a list."
            raise ValueError(msg)
        # make sure to unwind actual iterators, or the just might get exhausted
        # at some point
        identifiers = [identifier for identifier in value]
        for identifier in identifiers:
            _warn_on_invalid_uri(identifier)
        self._identifiers = identifiers

    @property
    def alternate_code(self):
        """
        From the StationXML definition:
            A code used for display or association, alternate to the
            SEED-compliant code.
        """
        return self._alternate_code

    @alternate_code.setter
    def alternate_code(self, value):
        if value:
            self._alternate_code = value.strip()
        else:
            self._alternate_code = None

    @property
    def historical_code(self):
        """
        From the StationXML definition:
            A previously used code if different from the current code.
        """
        return self._historical_code

    @historical_code.setter
    def historical_code(self, value):
        if value:
            self._historical_code = value.strip()
        else:
            self._historical_code = None

    def copy(self):
        """
        Returns a deepcopy of the object.

        :rtype: same class as original object
        :return: Copy of current object.

        .. rubric:: Examples

        1. Create a station object and copy it

            >>> from obspy import read_inventory
            >>> sta = read_inventory()[0][0]
            >>> sta2 = sta.copy()

           The two objects are not the same:

            >>> sta is sta2
            False

           But they have equal data (before applying further processing):

            >>> sta == sta2
            True

        2. The following example shows how to make an alias but not copy the
           data. Any changes on ``st3`` would also change the contents of
           ``st``.

            >>> sta3 = sta
            >>> sta is sta3
            True
            >>> sta == sta3
            True
        """
        return copy.deepcopy(self)

    def is_active(self, time=None, starttime=None, endtime=None):
        """
        Checks if the item was active at some given point in time (`time`)
        and/or if it was active at some point during a certain time range
        (`starttime`, `endtime`).

        .. note::
            If none of the time constraints is specified the result will always
            be `True`.

        :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param time: Only include networks/stations/channels active at given
            point in time.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Only include networks/stations/channels active at or
            after given point in time (i.e. channels ending before given time
            will not be shown).
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: Only include networks/stations/channels active before
            or at given point in time (i.e. channels starting after given time
            will not be shown).
        :rtype: bool
        :returns: `True`/`False` depending on whether the item matches the
            specified time criteria.
        """
        if time is not None:
            if self.start_date is not None and time < self.start_date:
                return False
            if self.end_date is not None and time > self.end_date:
                return False
        if starttime is not None and self.end_date is not None:
            if starttime > self.end_date:
                return False
        if endtime is not None and self.start_date is not None:
            if endtime < self.start_date:
                return False

        return True


class DataAvailability(ComparingObject):
    """
    A description of time series data availability. This information should
    be considered transient and is primarily useful as a guide for
    generating time series data requests. The information for a
    DataAvailability (time) span may be specific to the time range used in a
    request that resulted in the document or limited to the availability of
    data within the request range. These details may or may not be
    retained when synchronizing metadata between data centers.
    Spans of data are represented by a start time, end time, number of segments
    contained in the span and maximum time tear within a certain span.

    :param start: Start of time extent
    :type start: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param end: End of time extent
    :type end: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param spans: Time spans with detail information
    :type spans: list of :class:`DataAvailabilitySpan`
    """
    def __init__(self, start=None, end=None, spans=None):
        start = start is not None and UTCDateTime(start)
        self.start = start
        end = end is not None and UTCDateTime(end)
        self.end = end
        self.spans = spans or []

    @property
    def spans(self):
        return self._spans

    @spans.setter
    def spans(self, value):
        msg = 'Data availability spans must be of DataAvailabilitySpan type.'
        try:
            for item in value:
                if not isinstance(item, DataAvailabilitySpan):
                    raise TypeError
        except TypeError:
            raise TypeError(msg)
        self._spans = value

    def __str__(self):
        if not self.spans:
            span_info = 'no time span information'
        else:
            span_info = '%d time spans with details' % len(self.spans)
        return "Data Availability from %s to %s, %s." % (self.start,
                                                         self.end, span_info)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))


class DataAvailabilitySpan(ComparingObject):
    """
    Data availability spans are represented by a start time, end time, number
    of segments contained in the span and maximum time tear within a certain
    span.

    :param start: Start of time span
    :type start: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param end: End of time span
    :type end: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param number_of_segments: The number of continuous time series segments
        contained in the specified time range. A value of 1 indicates that the
        time series is continuous from start to end.
    :type number_of_segments: int
    :param maximum_time_tear: The maximum time tear (gap or overlap) in seconds
        between time series segments in the specified range.
    :type maximum_time_tear: float
    """
    def __init__(self, start, end, number_of_segments, maximum_time_tear=None):
        self.start = UTCDateTime(start)
        self.end = UTCDateTime(end)
        self.number_of_segments = number_of_segments
        self.maximum_time_tear = maximum_time_tear

    def __str__(self):
        if self.maximum_time_tear is None:
            tear_info = 'maximum time tear not specified'
        elif abs(self.maximum_time_tear) < 0.1:
            tear_info = '%.6fs maximum time tear'
        elif abs(self.maximum_time_tear) < 2:
            tear_info = '%.3fs maximum time tear'
        elif abs(self.maximum_time_tear) < 10:
            tear_info = '%.1fs maximum time tear'
        else:
            tear_info = '%.0fs maximum time tear'
        return "Data Availability Span: %d segments from %s to %s, %s." % (
            self.number_of_segments, self.start, self.end, tear_info)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))


class Equipment(ComparingObject):
    """
    An object containing a detailed description of an equipment.
    """
    def __init__(self, type=None, description=None, manufacturer=None,
                 vendor=None, model=None, serial_number=None,
                 installation_date=None, removal_date=None,
                 calibration_dates=None, resource_id=None):
        """
        :type type: str
        :param type: The equipment type
        :type description: str
        :param description: Description of the equipment
        :type manufacturer: str
        :param manufacturer: The manufacturer of the equipment
        :type vendor: str
        :param vendor: The vendor of the equipment
        :type model: str
        :param model: The model of the equipment
        :type serial_number: str
        :param serial_number: The serial number of the equipment
        :type installation_date: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param installation_date: The installation date of the equipment
        :type removal_date: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param removal_date: The removal data of the equipment
        :type calibration_dates: list of
            :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param calibration_dates: A list with all calibration dates of the
            equipment.
        :type resource_id: str
        :param resource_id: This field contains a string that should serve as a
            unique resource identifier. This identifier can be interpreted
            differently depending on the data center/software that generated
            the document. Also, we recommend to use something like
            GENERATOR:Meaningful ID. As a common behavior equipment with the
            same ID should contain the same information/be derived from the
            same base instruments.
        """
        self.type = type
        self.description = description
        self.manufacturer = manufacturer
        self.vendor = vendor
        self.model = model
        self.serial_number = serial_number
        self.installation_date = installation_date
        self.removal_date = removal_date
        self.calibration_dates = calibration_dates or []
        self.resource_id = resource_id

        @property
        def installation_date(self):
            return self._installation_date

        @installation_date.setter
        def installation_date(self, value):
            if value is None or isinstance(value, UTCDateTime):
                self._installation_date = value
                return
            self._installation_date = UTCDateTime(value)

        @property
        def removal_date(self):
            return self._removal_date

        @removal_date.setter
        def removal_date(self, value):
            if value is None or isinstance(value, UTCDateTime):
                self._removal_date = value
                return
            self._removal_date = UTCDateTime(value)

    def __str__(self):
        ret = ("Equipment:\n"
               "\tType: {type}\n"
               "\tDescription: {description}\n"
               "\tManufacturer: {manufacturer}\n"
               "\tVendor: {vendor}\n"
               "\tModel: {model}\n"
               "\tSerial number: {serial_number}\n"
               "\tInstallation date: {installation_date}\n"
               "\tRemoval date: {removal_date}\n"
               "\tResource id: {resource_id}\n"
               "\tCalibration Dates:\n")
        for calib_date in self.calibration_dates:
            ret += "\t\t%s\n" % calib_date
        ret = ret.format(**self.__dict__)
        return ret

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))


class Operator(ComparingObject):
    """
    An operating agency and associated contact persons. If there are multiple
    operators, each one should be encapsulated within an Operator object. Since
    the Contact element is a generic type that represents any contact person,
    it also has its own optional Agency element.
    """
    @deprecated_keywords({"agencies": "agency"})
    def __init__(self, agency, contacts=None, website=None):
        """
        :type agency: str
        :param agency: The agency of the operator.
        :type contacts: list of :class:`Person`, optional
        :param contacts: One or more contact persons.
        :type website: str, optional
        :param website: The website.
        """
        self.agency = agency
        self.contacts = contacts or []
        self.website = website

    @property
    def agency(self):
        return self._agency

    @agency.setter
    def agency(self, value):
        # check if a list of agencies was provided, which is not supported
        # anymore (if we get a string, types of provided value and any index
        # will match)
        if not isinstance(value[0], type(value)):
            msg = ("Only a single agency can be assigned to Operator due to "
                   "the changes in StationXML 1.1. Subsequent agencies are "
                   "ignored.")
            warnings.warn(msg, ObsPyDeprecationWarning)
            value = value[0]
        self._agency = value

    @property
    def agencies(self):
        msg = ("Attribute 'agencies' (holding a list of strings as Agencies) "
               "is deprecated in favor of 'agency' which now holds a single "
               "string (following changes in StationXML 1.1) and might be "
               "removed in the future. Returning a list built up of the "
               "single agency or an empty list if agency is None.")
        warnings.warn(msg, ObsPyDeprecationWarning)
        if self.agency is not None:
            return [self.agency]
        return []

    @agencies.setter
    def agencies(self, value):
        msg = ("Attribute 'agencies' (holding a list of strings as Agencies) "
               "is deprecated in favor of 'agency' which now holds a single "
               "string (following changes in StationXML 1.1) and might be "
               "removed in the future. Setting 'agency' with first item in "
               "provided list.")
        warnings.warn(msg, ObsPyDeprecationWarning)
        if not hasattr(value, "__iter__") or len(value) < 1:
            msg = ("agencies needs to be iterable, e.g. a list, and contain "
                   "at least one entry.")
            raise ValueError(msg)
        self._agency = value[0]

    @property
    def contacts(self):
        return self._contacts

    @contacts.setter
    def contacts(self, value):
        if not hasattr(value, "__iter__"):
            msg = ("contacts needs to be iterable, e.g. a list.")
            raise ValueError(msg)
        self._contacts = value


class Person(ComparingObject):
    """
    From the StationXML definition:
        Representation of a person's contact information. A person can belong
        to multiple agencies and have multiple email addresses and phone
        numbers.
    """
    email_pattern = re.compile(r"[\w\.\-_]+@[\w\.\-_]+")

    def __init__(self, names=None, agencies=None, emails=None, phones=None):
        """
        :type names: list[str], optional
        :param names: Self-explanatory. Multiple names allowed.
        :type agencies: list[str], optional
        :param agencies: Self-explanatory. Multiple agencies allowed.
        :type emails: list[str], optional
        :param emails: Self-explanatory. Multiple emails allowed.
        :type phones: list[:class:`PhoneNumber`], optional
        :param phones: Self-explanatory. Multiple phone numbers allowed.
        """
        self.names = names or []
        self.agencies = agencies or []
        self.emails = emails or []
        self.phones = phones or []

    @property
    def names(self):
        return self._names

    @names.setter
    def names(self, value):
        if not hasattr(value, "__iter__"):
            msg = "names needs to be iterable, e.g. a list."
            raise ValueError(msg)
        self._names = value

    @property
    def agencies(self):
        return self._agencies

    @agencies.setter
    def agencies(self, value):
        if not hasattr(value, "__iter__"):
            msg = "agencies needs to be iterable, e.g. a list."
            raise ValueError(msg)
        self._agencies = value

    @property
    def emails(self):
        return self._emails

    @emails.setter
    def emails(self, values):
        if not hasattr(values, "__iter__"):
            msg = "emails needs to be iterable, e.g. a list."
            raise ValueError(msg)
        for value in values:
            if re.match(self.email_pattern, value) is None:
                msg = ("emails needs to match the pattern "
                       r"'[\w\.\-_]+@[\w\.\-_]+'")
                raise ValueError(msg)
        self._emails = values

    @property
    def phones(self):
        return self._phones

    @phones.setter
    def phones(self, values):
        if not hasattr(values, "__iter__"):
            msg = "phones needs to be iterable, e.g. a list."
            raise ValueError(msg)
        self._phones = values


class PhoneNumber(ComparingObject):
    """
    A simple object representing a phone number.
    """
    phone_pattern = re.compile("^[0-9]+-[0-9]+$")

    def __init__(self, area_code, phone_number, country_code=None,
                 description=None):
        """
        :type area_code: int
        :param area_code: The area code.
        :type phone_number: str
        :param phone_number: The phone number minus the country and area code.
            Must be in the form "[0-9]+-[0-9]+", e.g. 1234-5678.
        :type country_code: int, optional
        :param country_code: The country code.
        :type description: str, optional
        :param description: Any additional information.
        """
        self.country_code = country_code
        self.area_code = area_code
        self.phone_number = phone_number
        self.description = description

    @property
    def phone_number(self):
        return self._phone_number

    @phone_number.setter
    def phone_number(self, value):
        if re.match(self.phone_pattern, value) is None:
            msg = "phone_number needs to match the pattern '[0-9]+-[0-9]+'"
            raise ValueError(msg)
        self._phone_number = value


class ExternalReference(ComparingObject):
    """
    From the StationXML definition:
        This type contains a URI and description for external data that users
        may want to reference in StationXML.
    """
    def __init__(self, uri, description):
        """
        :type uri: str
        :param uri: The URI to the external data.
        :type description: str
        :param description: A description of the external data.
        """
        self.uri = uri
        self.description = description


class Comment(ComparingObject):
    """
    From the StationXML definition:
        Container for a comment or log entry. Corresponds to SEED blockettes
        31, 51 and 59.
    """
    def __init__(self, value, id=None, begin_effective_time=None,
                 end_effective_time=None, authors=None, subject=None):
        """
        :type value: str
        :param value: The actual comment string
        :type id: int
        :param id: ID of comment, must be 0 or greater.
        :type begin_effective_time:
            :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        :param begin_effective_time: The effective start date.
        :type end_effective_time:
            :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        :param end_effective_time: The effective end date.
        :type authors: list of :class:`Person`, optional
        :param authors: The authors of this comment.
        :type subject: str, optional
        :param subject: Subject for relating comment, optional
        """
        self.value = value
        self.begin_effective_time = begin_effective_time
        self.end_effective_time = end_effective_time
        self.authors = authors or []
        self.id = id
        self.subject = subject

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value is None:
            self._id = value
            return
        if not int(value) >= 0:
            msg = "ID must be 0 or positive integer."
            raise ValueError(msg)
        self._id = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = str(value)

    @property
    def begin_effective_time(self):
        return self._begin_effective_time

    @begin_effective_time.setter
    def begin_effective_time(self, value):
        if value is None:
            self._begin_effective_time = None
            return
        self._begin_effective_time = UTCDateTime(value)

    @property
    def end_effective_time(self):
        return self._end_effective_time

    @end_effective_time.setter
    def end_effective_time(self, value):
        if value is None:
            self._end_effective_time = None
            return
        self._end_effective_time = UTCDateTime(value)

    @property
    def authors(self):
        return self._authors

    @authors.setter
    def authors(self, values):
        if not hasattr(values, "__iter__"):
            msg = "authors needs to be iterable, e.g. a list."
            raise ValueError(msg)
        self._authors = values

    def __str__(self):
        ret = ("Comment:\t{value}\n"
               "\tBegin Effective Time:\t{begin_effective_time}\n"
               "\tEnd Effective Time:\t{end_effective_time}\n"
               "\tAuthors:\t\t{authors}\n"
               "\tId:\t\t\t{id}")
        ret = ret.format(
            value=self.value, begin_effective_time=self.begin_effective_time,
            end_effective_time=self.end_effective_time, authors=self.authors,
            id=self.id)
        return ret

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))


class Site(ComparingObject):
    """
    From the StationXML definition:
        Description of a site location using name and optional geopolitical
        boundaries (country, city, etc.).
    """
    def __init__(self, name="", description=None, town=None, county=None,
                 region=None, country=None):
        """
        :type name: str
        :param name: The commonly used name of this station, equivalent to the
            SEED blockette 50, field 9.
        :type description: str, optional
        :param description: A longer description of the location of this
            station, e.g.  "NW corner of Yellowstone National Park" or "20
            miles west of Highway 40."
        :type town: str, optional
        :param town: The town or city closest to the station.
        :type county: str, optional
        :param county: The county.
        :type region: str, optional
        :param region: The state, province, or region of this site.
        :type country: str, optional
        :param country: The country.
        """
        self.name = name
        self.description = description
        self.town = town
        self.county = county
        self.region = region
        self.country = country

    def __str__(self):
        ret = ("Site: {name}\n"
               "\tDescription: {description}\n"
               "\tTown:    {town}\n"
               "\tCounty:  {county}\n"
               "\tRegion:  {region}\n"
               "\tCountry: {country}")
        ret = ret.format(
            name=self.name, description=self.description,
            town=self.town, county=self.county, region=self.region,
            country=self.country)
        return ret

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))


class Latitude(FloatWithUncertaintiesFixedUnit):
    """
    Latitude object

    :type value: float
    :param value: Latitude value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type datum: str
    :param datum: Datum for latitude coordinate
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = -90
    _maximum = 90
    _unit = "DEGREES"

    def __init__(self, value, lower_uncertainty=None, upper_uncertainty=None,
                 datum=None):
        """
        """
        self.datum = datum
        super(Latitude, self).__init__(
            value, lower_uncertainty=lower_uncertainty,
            upper_uncertainty=upper_uncertainty)


class Longitude(FloatWithUncertaintiesFixedUnit):
    """
    Longitude object

    :type value: float
    :param value: Longitude value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type datum: str
    :param datum: Datum for longitude coordinate
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = -180
    _maximum = 180
    unit = "DEGREES"

    def __init__(self, value, lower_uncertainty=None, upper_uncertainty=None,
                 datum=None):
        """
        """
        self.datum = datum
        super(Longitude, self).__init__(
            value, lower_uncertainty=lower_uncertainty,
            upper_uncertainty=upper_uncertainty)


class Distance(FloatWithUncertaintiesAndUnit):
    """
    Distance object

    :type value: float
    :param value: Distance value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type unit: str
    :param unit: Unit for distance measure.
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    def __init__(self, value, lower_uncertainty=None, upper_uncertainty=None,
                 unit="METERS"):
        super(Distance, self).__init__(
            value, lower_uncertainty=lower_uncertainty,
            upper_uncertainty=upper_uncertainty)
        self._unit = unit


class Azimuth(FloatWithUncertaintiesFixedUnit):
    """
    Azimuth object

    :type value: float
    :param value: Azimuth value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = 0
    _maximum = 360
    unit = "DEGREES"


class Dip(FloatWithUncertaintiesFixedUnit):
    """
    Dip object

    :type value: float
    :param value: Dip value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = -90
    _maximum = 90
    unit = "DEGREES"


class ClockDrift(FloatWithUncertaintiesFixedUnit):
    """
    ClockDrift object

    :type value: float
    :param value: ClockDrift value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = 0
    unit = "SECONDS/SAMPLE"


class SampleRate(FloatWithUncertaintiesFixedUnit):
    """
    SampleRate object

    :type value: float
    :param value: ClockDrift value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    unit = "SAMPLES/S"


class Frequency(FloatWithUncertaintiesFixedUnit):
    """
    Frequency object

    :type value: float
    :param value: Frequency value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    unit = "HERTZ"


class Angle(FloatWithUncertaintiesFixedUnit):
    """
    Angle object

    :type value: float
    :param value: Angle value
    :type lower_uncertainty: float
    :param lower_uncertainty: Lower uncertainty (aka minusError)
    :type upper_uncertainty: float
    :param upper_uncertainty: Upper uncertainty (aka plusError)
    :type measurement_method: str
    :param measurement_method: Method used in the measurement.
    """
    _minimum = -360
    _maximum = 360
    unit = "DEGREES"


def _unified_content_strings(contents):
    contents_unique = sorted(set(contents), key=_seed_id_keyfunction)
    contents_counts = [
        (item, contents.count(item)) for item in contents_unique]
    items = [item if count == 1 else "{} ({}x)".format(item, count)
             for item, count in contents_counts]
    return items


def _unified_content_strings_expanded(contents):
    contents2 = [["." + item.location_code, item.code,
                  item.sample_rate, item.start_date, item.end_date,
                  item.depth]
                 for item in contents]

    # sorts by startdate, sample rate, and channel code (ZNE321)
    contents2 = sorted(contents2, key=lambda x: (x[1], x[2], x[3]),
                       reverse=True)

    uniques = []
    for u in [[e[0], e[1][0:2], e[3], e[4], e[5]] for e in contents2]:
        if u not in uniques:
            uniques.append(u)

    contents3 = []
    for u in uniques:
        c = [e for e in contents2 if
             [e[0], e[1][0:2], e[3], e[4], e[5]] == u]
        test = [[e[0], e[2], e[3], e[4], e[5]] for e in c]
        if all(test[0] == x for x in test) and len(test) > 1:
            mergedch = u[1] + '[' \
                + ''.join(map(str, [e[1][-1] for e in c])) + ']'
            c[0][1] = mergedch
        contents3.append(c[0])

    contents3 = sorted(contents3, key=lambda x: (x[3], x[2], x[5]),
                       reverse=True)

    items = []
    for item in contents3:
        start_str = "%.10s(%03d)" % (str(item[3]),
                                     UTCDateTime(item[3]).julday)
        if item[4]:
            end_str = "%.10s(%03d)" % (str(item[4]),
                                       UTCDateTime(item[4]).julday)
        else:
            end_str = "    "  # or "None" ?
        if item[5]:
            items.append("{l: >5s}.{c: <9s}{sr: 6.1f} Hz  {start: <.15s}"
                         " - {end: <15.15s}  Depth {ldepth: >5.1f} m"
                         .format(l=item[0], c=item[1], sr=item[2],
                                 start=start_str, end=end_str,
                                 ldepth=item[5]))
        else:
            items.append("{l: >5s}.{c: <9s}{sr: 6.1f} Hz  {start: <.15s}"
                         " - {end: <.15s}"
                         .format(l=item[0], c=item[1], sr=item[2],
                                 start=start_str, end=end_str))

    return items


# make TextWrapper only split on colons, so that we avoid splitting in between
# e.g. network code and network code occurence count (can be controlled with
# class attributes).
# Also avoid lines starting with ", " (need to patch the class for this)
class InventoryTextWrapper(TextWrapper):
    wordsep_re = re.compile(r'(, )')
    wordsep_simple_re = re.compile(r'(, )')

    def _wrap_chunks(self, *args, **kwargs):
        """
        """
        # the following doesn't work somehow (likely because of future??)
        # lines = super(InventoryTextWrapper, self)._wrap_chunks(
        #     *args, **kwargs)
        lines = TextWrapper._wrap_chunks(self, *args, **kwargs)
        lines = [re.sub(r'([\b\s]+), (.*)', r'\1\2', line, count=1)
                 for line in lines]
        return lines


def _textwrap(text, *args, **kwargs):
    return InventoryTextWrapper(*args, **kwargs).wrap(text)


def _seed_id_keyfunction(x):
    """
    Keyfunction to use in sorting two (partial) SEED IDs

    Assumes that the last (or only) "."-separated part is a channel code.
    Assumes the last character is a the component code and sorts it
    "Z"-"N"-"E"-others_lexical.
    """
    # for comparison we build a list of 5 SEED code pieces:
    # [network, station, location, band+instrument, component]
    # with partial codes (i.e. not 4 fields after splitting at dots),
    # we go with the following assumptions (these seem a bit random, but that's
    # what can be encountered in string representations of the Inventory object
    # hierarchy):
    #  - no dot means network code only (e.g. "IU")
    #  - one dot means network.station code only (e.g. "IU.ANMO")
    #  - two dots means station.location.channel code only (e.g. "ANMO.10.BHZ")
    #  - three dots: full SEED ID (e.g. "IU.ANMO.10.BHZ")
    #  - more dots: sort after any of the previous, plain lexical sort
    # if no "." in the string: assume it's a network code

    # split to get rid of the description that that is added to networks and
    # stations which might also contain dots.
    number_of_dots = x.strip().split()[0].count(".")

    x = x.upper()
    if number_of_dots == 0:
        x = [x] + [""] * 4
    elif number_of_dots == 1:
        x = x.split(".") + [""] * 3
    elif number_of_dots in (2, 3):
        x = x.split(".")
        if number_of_dots == 2:
            x = [""] + x
        # split channel code into band+instrument code and component code
        x = x[:-1] + [x[-1][:-1], x[-1] and x[-1][-1] or '']
        # special comparison for component code, convert "ZNE" to integers
        # which compare less than any character
        comp = "ZNE".find(x[-1])
        # last item is component code, either the original 1-char string, or an
        # int from 0-2 if any of "ZNE". Python3 does not allow comparison of
        # int and string anymore (Python 2 always compares ints smaller than
        # any string), so we need to work around this by making this last item
        # a tuple with first item False for ints and True for strings.
        if comp >= 0:
            x[-1] = (False, comp)
        else:
            x[-1] = (True, x[-1])
    # all other cases, just convert the upper case string to a single item
    # list - it will compare greater than any of the split lists.
    else:
        x = [x, ]

    return x


def _response_plot_label(network, station, channel, label_epoch_dates):
    label = ".".join((network.code, station.code,
                      channel.location_code, channel.code))
    if label_epoch_dates:
        start = channel.start_date
        if start is None:
            start = 'open'
        else:
            start = str(start.date)
        end = channel.end_date
        if end is None:
            end = 'open'
        else:
            end = str(end.date)
        label += '\n{} -- {}'.format(start, end)
    return label


def _is_valid_uri(uri):
    if ':' not in uri:
        return False
    scheme, path = uri.split(':', 1)
    if any(not x.strip() for x in (scheme, path)):
        return False
    return True


def _warn_on_invalid_uri(uri):
    if not _is_valid_uri(uri):
        msg = f"Given string seems to not be a valid URI: '{uri}'"
        warnings.warn(msg)


def _add_resolve_seedid_doc(func):
    doc = """
    The following parameters deal with the problem, that the format
    only stores station names for the picks, but the Pick object expects
    a SEED id. The SEED id is looked up for every pick by the
    following procedure:

    1. look at seedid_map for a direct station name match and use the specified
       template
    2. if 1 did not succeed, look if the station is present in inventory and
       use its first channel as template
    3. if 1 and 2 did not succeed, use specified default template
       (default_seedid)

    :type inventory: :class:`~obspy.core.inventory.inventory.Inventory`
    :param inventory: Inventory used to retrieve network code, location code
        and channel code of stations (SEED id).
    :param dict seedid_map: Default templates for each station
        (example: `seedid_map={'MOX': 'GR.{}..HH{}'`).
        The values must contain three dots and two `{}` which are
        substituted by station code and component.
    :param str default_seedid: Default SEED id template.
        The value must contain three dots and two `{}` which are
        substituted by station code and component.
    :param bool warn: Whether or not to warn on failed look ups
       (no matching data found or ambiguous results) in the inventory
    """
    if func.__doc__ is not None:
        func.__doc__ = func.__doc__ + doc
    return func


def _add_resolve_seedid_ph2comp_doc(func):
    doc = """
    :param dict ph2comp: mapping of phases to components if format does not
        specify the component or if the component ends with '?'. Set it to
        `None` for no mapping of components. (default: {'P': 'Z', 'S': 'N'})
    """
    if func.__doc__ is not None:
        func.__doc__ = func.__doc__ + doc
    return func


def _resolve_seedid(station, component, inventory=None,
                    time=None, seedid_map=None, default_seedid=None,
                    key='{sta.code}', id_map=None, id_default=None,
                    phase=None, ph2comp={'P': 'Z', 'S': 'N'},
                    unused_kwargs=False, warn=True, **kwargs):
    if not unused_kwargs and len(kwargs) > 0:
        raise ValueError(f'Unexpected arguments: {kwargs}')
    if id_map is not None:  # backwards compatibility
        seedid_map = id_map
    if id_default is not None:  # backwards compatibility
        default_seedid = id_default
    if phase is not None and ph2comp is not None and (
            component == '' or component.endswith('?')):
        component = component[:-1] + ph2comp.get(phase.upper(), '')
    seedid = None
    if seedid_map is not None and station in seedid_map:
        seedid = seedid_map[station].format(station, component)
    elif inventory is not None:
        seedid = _resolve_seedid_from_inventory(
                station, component, inventory, time=time, warn=warn)
    if seedid is None and default_seedid is not None:
        seedid = default_seedid.format(station, component)
    if seedid is None:
        return '', station, None, component
    else:
        return tuple(seedid.split('.'))


def _resolve_seedid_from_inventory(
        station, component, inventory, time=None, network=None,
        location=None, warn=True):
    """
    Return a (Network, Station, Location, Channel) tuple.

    Given a station and channel code and station metadata (and optionally a
    certain point in time), try to resolve the full SEED ID, i.e. fill in
    a missing/unknown network and/or location code.
    If no matching data is found in metadata or if ambiguities in the station
    metadata are encountered, returns ``None`` for network and/or location
    code.

    Simply returns the given (Network, Station, Location, Channel) input if
    *both* ``location`` and ``network`` are already specified.

    :type station: str
    :param station: Station code to look up.
    :type channel: str
    :param channel: Channel code to look up.
    :type inventory: :class:`~obspy.core.inventory.inventory.Inventory`
    :param inventory: Station metadata to use for look up of missing parts of
        the full SEED ID.
    :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param time: Optionally restrict lookup from metadata to given timestamp.
    :type network: str
    :param network: Also specify network code for lookup (not intended to be
        used together with ``location``, see above)
    :type location: str
    :param location: Also specify location code for lookup (not intended to be
        used together with ``network``, see above)
    :type warn: bool
    :param warn: Whether or not to warn on failed look ups (no matching data
        found or ambiguous results) that return some ``None``s.
    :rtype: str
    :returns: SEED id string
    """
    inv = inventory.select(station=station, channel='*' + component, time=time,
                           network=network, location=location,
                           keep_empty=False)
    if len(inv.networks) != 1 or len(inv.networks[0].stations) == 0:
        if warn:
            msg = ('No matching metadata found for station '
                   f'{station}, component {component}.')
            warnings.warn(msg)
        return
    net = inv.networks[0]
    seedids = [f'{net.code}.{station}.{cha.location_code}.{cha.code}'
               for cha in net.stations[0] if cha.is_active(time=time)]
    seedids = [id_[:len(id_) - len(component)] + component for id_ in seedids]
    if len(seedids) == 0:
        if warn:
            msg = ('No matching metadata found for station '
                   f'{station}, component {component}.')
            warnings.warn(msg)
        return
    if len(set(seedids)) > 1 and warn:
        msg = ('Multiple SEED ids found for station '
               f'{station}, component {component}. Use first.')
        warnings.warn(msg)
    return seedids.pop(0)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.io.seisan - SEISAN read support for ObsPy
===============================================

The obspy.io.seisan package contains methods in order to read seismogram
files in the SEISAN format.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Reading
-------
Importing SEISAN files is done similar to reading any other waveform data
format within ObsPy by using the :func:`~obspy.core.stream.read()` method of
the :mod:`obspy.core` module. Example seismogram files may be found at
https://examples.obspy.org.

>>> from obspy import read
>>> st = read("/path/to/2001-01-13-1742-24S.KONO__004")
>>> st  # doctest: +ELLIPSIS
<obspy.core.stream.Stream object at 0x...>
>>> print(st)  # doctest: +ELLIPSIS
4 Trace(s) in Stream:
.KONO.0.B0Z | 2001-01-13T17:45:01.999000Z - ... | 20.0 Hz, 6000 samples
.KONO.0.L0Z | 2001-01-13T17:42:24.924000Z - ... | 1.0 Hz, 3542 samples
.KONO.0.L0N | 2001-01-13T17:42:24.924000Z - ... | 1.0 Hz, 3542 samples
.KONO.0.L0E | 2001-01-13T17:42:24.924000Z - ... | 1.0 Hz, 3542 samples

The file format will be determined automatically. Each trace (multiple channels
are mapped to multiple traces) will have a stats attribute containing the usual
information.

>>> print(st[0].stats)  # doctest: +NORMALIZE_WHITESPACE
             network:
             station: KONO
            location: 0
             channel: B0Z
           starttime: 2001-01-13T17:45:01.999000Z
             endtime: 2001-01-13T17:50:01.949000Z
       sampling_rate: 20.0
               delta: 0.05
                npts: 6000
               calib: 1.0
             _format: SEISAN

The actual data is stored as numpy.ndarray in the data attribute of each trace.

>>> print(st[0].data)
[  464   492   519 ..., -7042 -6960 -6858]
"""
if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



from __future__ import print_function

from math import log10

import numpy as np

from obspy.clients.arclink import Client
from obspy import UTCDateTime
from obspy.geodetics import gps2dist_azimuth


paz_wa = {'sensitivity': 2800, 'zeros': [0j], 'gain': 1,
          'poles': [-6.2832 - 4.7124j, -6.2832 + 4.7124j]}

client = Client(user="sed-workshop@obspy.org")
t = UTCDateTime("2012-04-03T02:45:03")

stations = client.get_stations(t, t + 300, "CH")
mags = []

for station in stations:
    station = station['code']
    try:
        st = client.get_waveforms("CH", station, "", "[EH]H[ZNE]", t - 300,
                                  t + 300, metadata=True)
        assert len(st) == 3
    except Exception:
        print(station, "---")
        continue

    st.simulate(paz_remove="self", paz_simulate=paz_wa, water_level=10)
    st.trim(t, t + 50)

    tr_n = st.select(component="N")[0]
    ampl_n = max(abs(tr_n.data))
    tr_e = st.select(component="E")[0]
    ampl_e = max(abs(tr_e.data))
    ampl = max(ampl_n, ampl_e)

    sta_lat = st[0].stats.coordinates.latitude
    sta_lon = st[0].stats.coordinates.longitude
    event_lat = 46.218
    event_lon = 7.706

    epi_dist, az, baz = gps2dist_azimuth(event_lat, event_lon, sta_lat,
                                         sta_lon)
    epi_dist = epi_dist / 1000

    if epi_dist < 60:
        a = 0.018
        b = 2.17
    else:
        a = 0.0038
        b = 3.02
    ml = log10(ampl * 1000) + a * epi_dist + b
    print(station, ml)
    mags.append(ml)

net_mag = np.median(mags)
print("Network magnitude:", net_mag)



# -*- coding: utf-8 -*-
from .blockette import Blockette
from ..fields import FixedString, Float, Integer, Loop, VariableString
from ..utils import blockette_34_lookup, format_resp


class Blockette043(Blockette):
    """
    Blockette 043: Response (Poles & Zeros) Dictionary Blockette.

    See Response (Poles & Zeros) Blockette [53] for more information.
    """

    id = 43
    name = "Response Poles and Zeros Dictionary"
    fields = [
        Integer(3, "Response Lookup Key", 4),
        VariableString(4, "Response Name", 1, 25, 'UN_'),
        FixedString(5, "Response type", 1, 'U'),
        Integer(6, "Stage signal input units", 3, xpath=34),
        Integer(7, "Stage signal output units", 3, xpath=34),
        Float(8, "A0 normalization factor", 12, mask='%+1.5e'),
        Float(9, "Normalization frequency", 12, mask='%+1.5e'),
        Integer(10, "Number of complex zeros", 3),
        # REPEAT fields 11 — 14 for the Number of complex zeros:
        Loop('Complex zero', "Number of complex zeros", [
            Float(11, "Real zero", 12, mask='%+1.5e'),
            Float(12, "Imaginary zero", 12, mask='%+1.5e'),
            Float(13, "Real zero error", 12, mask='%+1.5e'),
            Float(14, "Imaginary zero error", 12, mask='%+1.5e')
        ]),
        Integer(15, "Number of complex poles", 3),
        # REPEAT fields 16 — 19 for the Number of complex poles:
        Loop('Complex pole', "Number of complex poles", [
            Float(16, "Real pole", 12, mask='%+1.5e'),
            Float(17, "Imaginary pole", 12, mask='%+1.5e'),
            Float(18, "Real pole error", 12, mask='%+1.5e'),
            Float(19, "Imaginary pole error", 12, mask='%+1.5e')
        ])
    ]

# Changes the name of the blockette because of an error in XSEED 1.0
    def get_xml(self, *args, **kwargs):
        xml = Blockette.get_xml(self, *args, **kwargs)
        if self.xseed_version == '1.0':
            xml.tag = 'response_poles_and_zeros'
        return xml

    def get_resp(self, station, channel, abbreviations):
        """
        Returns RESP string.
        """
        # Field five needs some extra parsing.
        field_five_dict = {'A': 'A [Laplace Transform (Rad/sec)]',
                           'B': 'B [Analog (Hz)]',
                           'C': 'C [Composite]',
                           'D': 'D [Digital (Z-transform)]'}
        string = \
            '#\t\t+               ' + \
            '+-----------------------------------------' + \
            '---+                +\n' + \
            '#\t\t+               |   Response (Poles & Zeros),' + \
            '%6s ch %s   |                +\n' % (station, channel) + \
            '#\t\t+               ' + \
            '+-----------------------------------------' + \
            '---+                +\n' + \
            '#\t\t\n' + \
            'B043F05     Response type:                         %s\n' \
            % field_five_dict[self.response_type] + \
            'B043F06     Response in units lookup:              %s\n' \
            % blockette_34_lookup(abbreviations,
                                  self.stage_signal_input_units) + \
            'B043F07     Response out units lookup:             %s\n' \
            % blockette_34_lookup(abbreviations,
                                  self.stage_signal_output_units) + \
            'B043F08     A0 normalization factor:               %G\n'\
            % self.A0_normalization_factor + \
            'B043F09     Normalization frequency:               %G\n'\
            % self.normalization_frequency + \
            'B043F10     Number of zeroes:                      %s\n'\
            % self.number_of_complex_zeros + \
            'B043F15     Number of poles:                       %s\n'\
            % self.number_of_complex_poles + \
            '#\t\tComplex zeroes:\n' + \
            '#\t\t  i  real          imag          real_error    imag_error\n'
        if self.number_of_complex_zeros > 0:
            if self.number_of_complex_zeros != 1:
                # Loop over all zeros.
                for _i in range(self.number_of_complex_zeros):
                    string += 'B043F11-14 %4s %13s %13s %13s %13s\n' % (
                        _i,
                        format_resp(self.real_zero[_i], 6),
                        format_resp(self.imaginary_zero[_i], 6),
                        format_resp(self.real_zero_error[_i], 6),
                        format_resp(self.imaginary_zero_error[_i], 6))
            else:
                string += 'B043F11-14 %4s %13s %13s %13s %13s\n' % (
                    0,
                    format_resp(self.real_zero, 6),
                    format_resp(self.imaginary_zero, 6),
                    format_resp(self.real_zero_error, 6),
                    format_resp(self.imaginary_zero_error, 6))
        string += '#\t\tComplex poles:\n' + \
            '#\t\t  i  real          imag          real_error    imag_error\n'
        if self.number_of_complex_poles > 0:
            if self.number_of_complex_poles != 1:
                # Loop over all poles.
                for _i in range(self.number_of_complex_poles):
                    string += 'B043F16-19 %4s %13s %13s %13s %13s\n' % (
                        _i,
                        format_resp(self.real_pole[_i], 6),
                        format_resp(self.imaginary_pole[_i], 6),
                        format_resp(self.real_pole_error[_i], 6),
                        format_resp(self.imaginary_pole_error[_i], 6))
            else:
                string += 'B043F16-19 %4s %13s %13s %13s %13s\n' % (
                    0,
                    format_resp(self.real_pole, 6),
                    format_resp(self.imaginary_pole, 6),
                    format_resp(self.real_pole_error, 6),
                    format_resp(self.imaginary_pole_error, 6))
        string += '#\t\t\n'
        return string.encode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Domain definitions for the download helpers.

Subclass the :class:`~obspy.clients.fdsn.mass_downloader.domain.Domain` class
to define custom and potentially more complex domains. See its documentation
for an example.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2014-2015
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from abc import ABCMeta, abstractmethod


class Domain(metaclass=ABCMeta):
    """
    Abstract base class defining a domain - subclass it to define a new domain.

    Each subclass must implement the :meth:`~.get_query_parameters`
    method and optionally the :meth:`~.is_in_domain` method which enables
    the construction of arbitrarily complex domains. The
    :meth:`~.get_query_parameters` method must return the query parameters to
    download as much data as required. The :meth:`~.is_in_domain` can later be
    used to refine the domain after the data has been downloaded.


    It can be thought of as a boolean operation - first the rough domain is
    specified including all the possible points, then some points are
    removed again. This is illustrated with an example domain representing
    Germany utilizing the external packages
    `shapely <http://toblerity.org/shapely/>`_ and
    `fiona <http://toblerity.org/fiona/>`_. Shapefiles can be found online
    on many websites, for example on
    `http://www.gadm.org/ <http://www.gadm.org/>`_. The example works by
    first extracting the bounds of the country to formulate the FDSN query
    and then removing points outside of the exact shape.


    .. code-block:: python

        import fiona
        import shapely.geometry

        from obspy.clients.fdsn.mass_downloader import Domain


        class Germany(Domain):
            def __init__(self):
                Domain.__init__(self)

                fiona_collection = fiona.open("./DEU_adm/DEU_adm0.shp")
                geometry = fiona_collection.next()["geometry"]

                self.shape = shapely.geometry.asShape(geometry)
                self.b = fiona_collection.bounds

            def get_query_parameters(self):
                return {"minlatitude": self.b[1],
                        "minlongitude": self.b[0],
                        "maxlatitude": self.b[3],
                        "maxlongitude": self.b[2]}

            def is_in_domain(self, latitude, longitude):
                if self.shape.contains(shapely.geometry.Point(longitude,
                                                              latitude)):
                    return True
                return False


    This is further illustrated by the following image. The green rectangle
    denotes the original FDSN query which returns the blue points. In the
    second step the red points are discarded leaving only points (stations)
    within Germany.

    .. figure:: /_images/expensive_plots/mass_downloader_domain.png
        :align: center

    """
    @abstractmethod
    def get_query_parameters(self):
        """
        Return the domain specific query parameters for the
        :meth:`~obspy.clients.fdsn.client.Client.get_stations` method as a
        dictionary. Possibilities keys for rectangular queries are

        * ``minlatitude``
        * ``maxlatitude``
        * ``minlongitude``
        * ``maxlongitude``

        For circular queries:

        * ``latitude``
        * ``longitude``
        * ``minradius``
        * ``maxradius``
        """
        pass

    def is_in_domain(self, latitude, longitude):
        """
        Returns True/False depending on the point being in the domain.

        If not implemented no further restrictions will be applied after the
        data has been downloaded.
        """
        raise NotImplementedError


class RectangularDomain(Domain):
    """
    A rectangular domain defined by latitude and longitude bounds.

    >>> domain = RectangularDomain(minlatitude=30, maxlatitude=50,
    ...                            minlongitude=5, maxlongitude=35)

    """
    def __init__(self, minlatitude, maxlatitude, minlongitude,
                 maxlongitude):
        self.minlatitude = minlatitude
        self.maxlatitude = maxlatitude
        self.minlongitude = minlongitude
        self.maxlongitude = maxlongitude

    def get_query_parameters(self):
        return {
            "minlatitude": self.minlatitude,
            "maxlatitude": self.maxlatitude,
            "minlongitude": self.minlongitude,
            "maxlongitude": self.maxlongitude}


class CircularDomain(Domain):
    """
    A circular domain defined by a center point and minimum and maximum
    radius from that point in degrees.

    >>> domain = CircularDomain(latitude=37.52, longitude=143.04,
    ...                         minradius=70.0, maxradius=90.0)

    """
    def __init__(self, latitude, longitude, minradius, maxradius):
        self.latitude = latitude
        self.longitude = longitude
        self.minradius = minradius
        self.maxradius = maxradius

    def get_query_parameters(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "minradius": self.minradius,
            "maxradius": self.maxradius}


class GlobalDomain(Domain):
    """
    Domain spanning the whole globe.

    >>> domain = GlobalDomain()
    """
    def get_query_parameters(self):
        return {}


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



import obspy


threechannels = obspy.read('https://examples.obspy.org/COP.BHE.DK.2009.050')
threechannels += obspy.read('https://examples.obspy.org/COP.BHN.DK.2009.050')
threechannels += obspy.read('https://examples.obspy.org/COP.BHZ.DK.2009.050')
threechannels.plot(size=(800, 600))



import numpy as np
import matplotlib.pyplot as plt

import obspy
from obspy.imaging.cm import obspy_sequential
from obspy.signal.tf_misfit import cwt


st = obspy.read()
tr = st[0]
npts = tr.stats.npts
dt = tr.stats.delta
t = np.linspace(0, dt * npts, npts)
f_min = 1
f_max = 50

scalogram = cwt(tr.data, dt, 8, f_min, f_max)

fig = plt.figure()
ax = fig.add_subplot(111)

x, y = np.meshgrid(
    t,
    np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]))

ax.pcolormesh(x, y, np.abs(scalogram), cmap=obspy_sequential)
ax.set_xlabel("Time after %s [s]" % tr.stats.starttime)
ax.set_ylabel("Frequency [Hz]")
ax.set_yscale('log')
ax.set_ylim(f_min, f_max)
plt.show()
